#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import asyncio
import datetime
from typing import TYPE_CHECKING, Iterable, Optional, List, overload, Callable, TypeVar, Type, Tuple, Union

from . import abc, utils
from .enum import ChannelType, try_enum
from .error import ClientException
from .mixins import Hashable
from .object import Object
from .schedule import Schedule

__all__ = (
    'TextChannel',
    'VoiceChannel',
    'LiveChannel',
    'AppChannel',
    'CategoryChannel',
    'ThreadChannel',
    'PartialMessageable',
    'DMChannel'
)

from .thread import Thread

from .utils import MISSING

if TYPE_CHECKING:
    from .user import *
    from .member import Member
    from .message import Message, PartialMessage
    from .state import ConnectionState
    from .guild import Guild, GuildChannel as GuildChannelType
    from .types.channel import (
        TextChannel as TextChannelPayload,
        VoiceChannel as VoiceChannelPayload,
        CategoryChannel as CategoryChannelPayload,
        LiveChannel as LiveChannelPayload,
        AppChannel as AppChannelPayload,
        ThreadChannel as ThreadChannelPayload,
        DMChannel as DMChannelPayload,
    )


async def _single_delete_strategy(messages: Iterable[Message]):
    for m in messages:
        await m.delete()


class TextChannel(abc.Messageable, abc.GuildChannel, Hashable):
    """代表 QQ 频道中的文字子频道

    .. container:: operations
        .. describe:: x == y

            检查两个子频道是否相等。

        .. describe:: x != y

            检查两个子频道是否不相等。

        .. describe:: hash(x)

            返回子频道的哈希值。

        .. describe:: str(x)

            返回子频道的名称。

    Attributes
    -----------
    name: :class:`str`
        子频道的名称。
    guild: :class:`Guild`
        子频道所属的频道。
    id: :class:`int`
        子频道 ID。
    private_type: :class:`int`
        子频道私密类型。

        +----+---------------------+
        | 值 | 含义                |
        +----+---------------------+
        | 0  | 公开频道            |
        +----+---------------------+
        | 1  | 群主管理员可见      |
        +----+---------------------+
        | 2  | 群主管理员+指定成员 |
        +----+---------------------+

    category_id: Optional[:class:`int`]
        这个子频道属于的分类频道，如果没有则返回 ``None`` 。
    position: :class:`int`
        在子频道列表中的位置。 这是一个从 0 开始的数字。例如顶部子频道是位置 0。
    last_message_id: Optional[:class:`int`]
        发送到此通道的消息的最后一个消息 ID。 它可能 *不* 指向现有的或有效的消息。

    """

    __slots__ = (
        'name',
        'id',
        'guild',
        '_state',
        'category_id',
        'position',
        '_type',
        'last_message_id',
        'private_type'
    )

    def __init__(self, *, state: ConnectionState, guild: Guild, data: TextChannelPayload):
        self._state: ConnectionState = state
        self.id: int = int(data['id'])
        self._type: int = data['type']
        self.last_message_id: Optional[int] = None
        self._update(guild, data)

    def __repr__(self) -> str:
        attrs = [
            ('id', self.id),
            ('name', self.name),
            ('position', self.position),
            ('category_id', self.category_id),
        ]
        joined = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {joined}>'

    def _update(self, guild: Guild, data: TextChannelPayload) -> None:
        self.guild: Guild = guild
        self.name: str = data['name']
        self.category_id: Optional[int] = int(data.get('parent_id')) if 'parent_id' in data else self.category_id
        self.position: int = data['position'] if 'position' in data else self.position
        self._type: int = data.get('type', self._type)
        self.private_type: int = data.get('private_type')

    async def _get_channel(self):
        return self, False

    @property
    def type(self) -> ChannelType:
        return try_enum(ChannelType, self._type)

    @property
    def _sorting_bucket(self) -> int:
        return ChannelType.text.value

    @property
    def members(self) -> List[Member]:
        """List[:class:`Member`]: 返回可以看到此子频道的所有成员。"""
        return [m for m in self.guild.members]

    def get_partial_message(self, message_id: int, /) -> PartialMessage:
        """从消息 ID 创建一个 :class:`PartialMessage`。
        如果你想处理消息并且只有它的 ID 而不进行不必要的 API 调用，这将非常有用。

        Parameters
        ------------
        message_id: :class:`int`
            要为其创建部分消息的消息 ID。
        Returns
        ---------
        :class:`PartialMessage`
            部分消息。
        """

        from .message import PartialMessage

        return PartialMessage(channel=self, id=message_id)

    @overload
    async def edit(
            self,
            *,
            reason: Optional[str] = ...,
            name: str = ...,
            position: int = ...,
            category: Optional[CategoryChannel] = ...,
            type: ChannelType = ...,
    ) -> Optional[TextChannel]:
        ...

    @overload
    async def edit(self) -> Optional[TextChannel]:
        ...

    async def edit(self, *, reason=None, **options):
        """|coro|
        编辑子频道。

        Parameters
        ----------
        name: :class:`str`
            新频道名称。
        position: :class:`int`
            新频道的位置。
        category: Optional[:class:`CategoryChannel`]
            此频道的新类别。可以是 ``None`` 以删除类别。
        type: :class:`ChannelType`
            更改此文本频道的类型。 不保证能够成功转换。
        reason: Optional[:class:`str`]
            编辑此频道的原因。显示在审计日志中。

        Raises
        ------
        InvalidArgument
            如果 position 小于 0 或大于子频道数。
        Forbidden
            你没有编辑频道的权限。
        HTTPException
            编辑频道失败。
        Returns
        --------
        Optional[:class:`.TextChannel`]
            新编辑的文本通道。如果编辑只是位置性的，则返回 ``None`` 。
        """

        payload = await self._edit(options, reason=reason)
        if payload is not None:
            # the payload will always be the proper channel payload
            return self.__class__(state=self._state, guild=self.guild, data=payload)  # type: ignore

    @utils.copy_doc(abc.GuildChannel.clone)
    async def clone(self, *, name: Optional[str] = None, reason: Optional[str] = None) -> TextChannel:
        return await self._clone_impl(
            {}, name=name, reason=reason
        )

    async def delete_messages(self, messages: Iterable[Message]) -> None:
        """|coro|
        删除消息列表。这类似于 :meth:`Message.delete`，除了它批量删除多条消息。
        作为一种特殊情况，如果消息数为 0，则什么也不做。如果消息数为 1，则完成单个消息删除。
        如果超过两个，则使用批量删除。你不能批量删除超过 100 条消息或超过 14 天的消息。

        Parameters
        -----------
        messages: Iterable[:class:`qq.Message`]
            一个可迭代的消息，表示要批量删除的消息。

        Raises
        ------
        ClientException
            要删除的消息数量超过 100 条。
        Forbidden
            你没有删除消息的适当权限。
        NotFound
            如果单次删除，则该消息已被删除。
        HTTPException
            删除消息失败。
        """
        if not isinstance(messages, (list, tuple)):
            messages = list(messages)

        if len(messages) == 0:
            return  # do nothing

        if len(messages) == 1:
            message_id: str = messages[0].id
            await self._state.http.delete_message(self.id, message_id)
            return

        if len(messages) > 100:
            raise ClientException('最多只能批量删除 100 条消息')

        message_ids = [m.id for m in messages]
        await self._state.http.delete_messages(self.id, message_ids)

    async def purge(
            self,
            *,
            limit: Optional[int] = 100,
            check: Callable[[Message], bool] = MISSING,
            before: Optional[datetime.datetime] = None,
            after: Optional[datetime.datetime] = None,
            around: Optional[datetime.datetime] = None,
            oldest_first: Optional[bool] = False,
            bulk: bool = True,
    ) -> List[Message]:
        """|coro|
        清除符合检查函数 ``check`` 给定标准的消息列表。
        如果未提供 ``check`` ，则所有消息都将被删除，一视同仁。

        Examples
        ---------
        删除机器人的消息: ::

            def is_me(m):
                return m.author == client.user
            deleted = await channel.purge(limit=100, check=is_me)
            await channel.send(f'已删除 {len(deleted)} 个消息')

        Parameters
        -----------
        limit: Optional[:class:`int`]
            要搜索的消息数。这不是将被删除的消息数，尽管可以是。
        check: Callable[[:class:`Message`], :class:`bool`]
            用于检查是否应删除消息的功能。它必须将 Message 作为其唯一参数。
        before: Optional[:class:`datetime.datetime`]
            与 :meth:`history` 中的 ``before`` 相同。
        after: Optional[:class:`datetime.datetime`]
            与 :meth:`history` 中的 ``after`` 相同。
        around: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
            与 :meth:`history`中的 ``around`` 相同。
        oldest_first: Optional[:class:`bool`]
            与 :meth:`history` 中的 ``oldest_first`` 相同。
        bulk: :class:`bool`
            如果 ``True`` ，使用批量删除。

        Raises
        -------
        Forbidden
            你没有执行所需操作的适当权限。
        HTTPException
            清除消息失败。

        Returns
        --------
        List[:class:`.Message`]
            已删除的消息列表。
        """

        if check is MISSING:
            check = lambda m: True

        iterator = self.history(limit=limit, before=before, after=after, oldest_first=oldest_first, around=around)
        ret: List[Message] = []
        count = 0
        strategy = self.delete_messages if bulk else _single_delete_strategy

        async for message in iterator:
            if count == 100:
                to_delete = ret[-100:]
                await strategy(to_delete)
                count = 0
                await asyncio.sleep(1)

            if not check(message):
                continue

            count += 1
            ret.append(message)

        # SOme messages remaining to poll
        if count >= 2:
            # more than 2 messages -> bulk delete
            to_delete = ret[-count:]
            await strategy(to_delete)
        elif count == 1:
            # delete a single message
            await ret[-1].delete()

        return ret

    async def unpin(self, reason: Optional[str] = None):
        """
        删除所有此子频道的公告。

        Parameters
        -----------
        reason: Optional[:class:`str`]
            删除公告的原因。

        Raises
        -------
        Forbidden
            你没有足够权限删除公告。
        NotFound
            消息或频道无法被找到，可能被删除了.
        HTTPException
            删除公告失败。
        """
        await self._state.http.channel_unpin_message(self.id, 'all', reason=reason)

    async def send_guide(self, content: str):
        """
        发送主动消息引导信息。

        Parameters
        -----------
        content: :class:`str`
            要发送的内容。

        Raises
        -------
        Forbidden
            你没有足够权限发送。
        HTTPException
            发送失败。
        """
        await self._state.http.send_guide(self.id, content)


class VoiceChannel(abc.GuildChannel, Hashable):
    """表示 QQ 语音子频道。

    .. container:: operations

        .. describe:: x == y

            检查两个子频道是否相等。

        .. describe:: x != y

            检查两个子频道是否不相等。

        .. describe:: hash(x)

            返回子频道的哈希值。

        .. describe:: str(x)

            返回子频道的名称。

    Attributes
    -----------
    name: :class:`str`
        子频道名称。
    guild: :class:`Guild`
        子频道所属的频道。
    id: :class:`int`
        频道 ID。
    private_type: :class:`int`
        子频道私密类型。

        +----+---------------------+
        | 值 | 含义                |
        +----+---------------------+
        | 0  | 公开频道            |
        +----+---------------------+
        | 1  | 群主管理员可见      |
        +----+---------------------+
        | 2  | 群主管理员+指定成员 |
        +----+---------------------+

    position: :class:`int`
        在类别列表中的位置。 这是一个从 0 开始的数字。例如顶部频道是位置 0。
    """
    __slots__ = (
        'name',
        'id',
        'guild',
        '_state',
        'position',
        'category_id',
        'private_type'
    )

    def __init__(self, *, state: ConnectionState, guild: Guild, data: VoiceChannelPayload):
        self._state: ConnectionState = state
        self.id: int = int(data['id'])
        self._update(guild, data)

    def _update(self, guild: Guild, data: VoiceChannelPayload) -> None:
        self.guild = guild
        self.name: str = data['name']
        self.category_id: Optional[int] = int(data.get('parent_id'))
        self.position: int = data['position']
        self.private_type: int = data.get('private_type')

    def __repr__(self) -> str:
        attrs = [
            ('id', self.id),
            ('name', self.name),
            ('position', self.position),
            ('category_id', self.category_id),
        ]
        joined = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {joined}>'

    @property
    def type(self) -> ChannelType:
        """:class:`ChannelType`: 子频道的QQ类型。"""
        return ChannelType.voice

    @utils.copy_doc(abc.GuildChannel.clone)
    async def clone(self, *, name: Optional[str] = None, reason: Optional[str] = None) -> VoiceChannel:
        return await self._clone_impl({}, name=name, reason=reason)

    @overload
    async def edit(
            self,
            *,
            name: str = ...,
            position: int = ...,
            category: Optional[CategoryChannel] = ...,
            reason: Optional[str] = ...,
    ) -> Optional[VoiceChannel]:
        ...

    @overload
    async def edit(self) -> Optional[VoiceChannel]:
        ...

    async def edit(self, *, reason=None, **options):
        """|coro|
        编辑频道。

        Parameters
        ----------
        name: :class:`str`
            新频道的名称。
        position: :class:`int`
            新频道的位置。
        category: Optional[:class:`CategoryChannel`]
            此频道的新类别。可以是 ``None`` 以删除类别。
        reason: Optional[:class:`str`]
            编辑此频道的原因。显示在审计日志中。

        Raises
        ------
        InvalidArgument
            如果权限覆盖信息格式不正确。
        Forbidden
            你没有编辑频道的权限。
        HTTPException
            编辑频道失败。

        Returns
        --------
        Optional[:class:`.VoiceChannel`]
            新编辑的语音通道。如果编辑只是位置性的，则返回 ``None`` 。
        """

        payload = await self._edit(options, reason=reason)
        if payload is not None:
            # the payload will always be the proper channel payload
            return self.__class__(state=self._state, guild=self.guild, data=payload)  # type: ignore


class LiveChannel(abc.GuildChannel, Hashable, abc.Messageable):
    """表示 QQ 直播子频道。

    .. container:: operations

        .. describe:: x == y

            检查两个子频道是否相等。

        .. describe:: x != y

            检查两个子频道是否不相等。

        .. describe:: hash(x)

            返回子频道的哈希值。

        .. describe:: str(x)

            返回子频道的名称。

    Attributes
    -----------
    name: :class:`str`
        子频道名称。
    guild: :class:`Guild`
        子频道所属的频道。
    id: :class:`int`
        频道 ID。
    private_type: :class:`int`
        子频道私密类型。

        +----+---------------------+
        | 值 | 含义                |
        +----+---------------------+
        | 0  | 公开频道            |
        +----+---------------------+
        | 1  | 群主管理员可见      |
        +----+---------------------+
        | 2  | 群主管理员+指定成员 |
        +----+---------------------+

    position: :class:`int`
        在类别列表中的位置。 这是一个从 0 开始的数字。例如顶部频道是位置 0。
    """

    __slots__ = (
        'name',
        'id',
        'guild',
        '_state',
        'position',
        'category_id',
        'private_type'
    )

    def __init__(self, *, state: ConnectionState, guild: Guild, data: LiveChannelPayload):
        self._state: ConnectionState = state
        self.id: int = int(data['id'])
        self._update(guild, data)

    def _update(self, guild: Guild, data: LiveChannelPayload) -> None:
        self.guild = guild
        self.name: str = data['name']
        self.category_id: Optional[int] = int(data.get('parent_id'))
        self.position: int = data['position']
        self.private_type: int = data.get('private_type')

    def __repr__(self) -> str:
        attrs = [
            ('id', self.id),
            ('name', self.name),
            ('position', self.position),
            ('category_id', self.category_id),
        ]
        joined = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {joined}>'


class AppChannel(abc.GuildChannel, Hashable):
    """表示 QQ 应用子频道。

    .. container:: operations

        .. describe:: x == y

            检查两个子频道是否相等。

        .. describe:: x != y

            检查两个子频道是否不相等。

        .. describe:: hash(x)

            返回子频道的哈希值。

        .. describe:: str(x)

            返回子频道的名称。

    Attributes
    -----------
    name: :class:`str`
        子频道名称。
    guild: :class:`Guild`
        子频道所属的频道。
    id: :class:`int`
        频道 ID。
    private_type: :class:`int`
        子频道私密类型。

        +----+---------------------+
        | 值 | 含义                |
        +----+---------------------+
        | 0  | 公开频道            |
        +----+---------------------+
        | 1  | 群主管理员可见      |
        +----+---------------------+
        | 2  | 群主管理员+指定成员 |
        +----+---------------------+

    position: :class:`int`
        在类别列表中的位置。 这是一个从 0 开始的数字。例如顶部频道是位置 0。
    """
    __slots__ = (
        'name',
        'id',
        'guild',
        '_state',
        'position',
        'category_id',
        'private_type'
    )

    def __init__(self, *, state: ConnectionState, guild: Guild, data: AppChannelPayload):
        self._state: ConnectionState = state
        self.id: int = int(data['id'])
        self._update(guild, data)

    def _update(self, guild: Guild, data: AppChannelPayload) -> None:
        self.guild = guild
        self.name: str = data['name']
        self.category_id: Optional[int] = int(data.get('parent_id'))
        self.position: int = data['position']
        self.private_type: int = data.get('private_type')

    def __repr__(self) -> str:
        attrs = [
            ('id', self.id),
            ('name', self.name),
            ('position', self.position),
            ('category_id', self.category_id),
        ]
        joined = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {joined}>'

    async def create_schedule(
            self,
            name: str,
            start: Union[datetime.datetime, float],
            end: Union[datetime.datetime, float],
            jump_channel: GuildChannelType,
            remind_type: str,
            description: Optional[str] = None,
    ) -> Schedule:
        """返回具有给定 path 和 method 的权限。

        Parameters
        -----------
        name: :class:`str`
            成员所属的频道。
        description: Optional[:class:`str`]
        start: Union[:class:`datetime.datetime`, :class:`float`]
            日程开始的时间。
        end: Union[:class:`datetime.datetime`, :class:`float`]
            日程开始的时间。
        jump_channel: :class:`GuildChannel`
            日程开始的时间。
        remind_type: :class:`str`
            日程提醒类型。

            +-----------+--------------+
            | 提醒类型 id	  | 描述           |
            +===========+==============+
            | 0         | 不提醒          |
            | 1         | 开始时提醒        |
            | 2         | 开始前 5 分钟提醒   |
            | 3         | 开始前 15 分钟提醒  |
            | 4         | 开始前 30 分钟提醒  |
            | 5         | 开始前 60 分钟提醒  |
            +-----------+--------------+


        Returns
        --------
        Optional[:class:`Schedule`]
            Role 或如果未找到，则  ``None`` 。
        """
        schedule = await self._state.http.create_schedule(
            channel_id=self.id,
            name=name,
            start_timestamp=start,
            end_timestamp=end,
            jump_channel_id=jump_channel.id,
            remind_type=remind_type,
            description=description
        )

        return Schedule(data=schedule, state=self._state, guild=self.guild, channel=self)


class ThreadChannel(abc.GuildChannel, Hashable):
    """表示 QQ 论坛子频道。

    .. container:: operations

        .. describe:: x == y

            检查两个子频道是否相等。

        .. describe:: x != y

            检查两个子频道是否不相等。

        .. describe:: hash(x)

            返回子频道的哈希值。

        .. describe:: str(x)

            返回子频道的名称。

    Attributes
    -----------
    name: :class:`str`
        子频道名称。
    guild: :class:`Guild`
        子频道所属的频道。
    id: :class:`int`
        频道 ID。
    private_type: :class:`int`
        子频道私密类型。

        +----+---------------------+
        | 值 | 含义                |
        +----+---------------------+
        | 0  | 公开频道            |
        +----+---------------------+
        | 1  | 群主管理员可见      |
        +----+---------------------+
        | 2  | 群主管理员+指定成员 |
        +----+---------------------+

    position: :class:`int`
        在类别列表中的位置。 这是一个从 0 开始的数字。例如顶部频道是位置 0。
    """
    __slots__ = (
        'name',
        'id',
        'guild',
        '_state',
        'position',
        'category_id',
        'private_type'
    )

    def __init__(self, *, state: ConnectionState, guild: Guild, data: ThreadChannelPayload):
        self._state: ConnectionState = state
        self.id: int = int(data['id'])
        self._update(guild, data)

    def _update(self, guild: Guild, data: ThreadChannelPayload) -> None:
        self.guild = guild
        self.name: str = data['name']
        self.category_id: Optional[int] = int(data.get('parent_id'))
        self.position: int = data['position']
        self.private_type: int = data.get('private_type')

    def __repr__(self) -> str:
        attrs = [
            ('id', self.id),
            ('name', self.name),
            ('position', self.position),
            ('category_id', self.category_id),
        ]
        joined = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {joined}>'

    async def fetch_threads(self) -> List[Thread]:
        """|coro|
        获取频道的所有帖子。

        Returns
        --------
        List[:class:`Thread`]
            频道的帖子列表，注意可能不完全。
        """
        threads = await self._state.http.get_thread_list(self.id)['threads']
        return [Thread(data=thread, state=self._state, channel=self) for thread in threads]

    async def create_thread(
            self,
            title: str,
            content: str,
            thread_type: int,
    ):
        """|coro|
        使用给定的内容向频道发送帖子。

        Parameters
        -----------
        title: :class:`str`
            帖子标题。
        content: :class:`str`
            帖子内容。
        thread_type: :class:`int`
            帖子类型。

            +---+------------------------------+
            | 值 | 描述                           |
            +===+==============================+
            | 1 | 普通文本                         |
            | 2 | HTML                         |
            | 3 | Markdown                     |
            | 4 | JSON（content参数可参照RichText结构） |
            +---+------------------------------+
        """
        await self._state.http.create_thread(
            channel_id=self.id,
            title=title,
            content=content,
            thread_type=thread_type
        )

    async def send_text(
            self,
            title: str,
            content: str,
    ):
        """|coro|
        使用给定的内容向频道发送普通文本帖子。

        Parameters
        -----------
        title: :class:`str`
            帖子标题。
        content: :class:`str`
            帖子内容。
        """
        await self.create_thread(title, content, 1)

    async def send_html(
            self,
            title: str,
            content: str,
    ):
        """|coro|
        使用给定的内容向频道发送HTML帖子。

        Parameters
        -----------
        title: :class:`str`
            帖子标题。
        content: :class:`str`
            帖子内容。
        """
        await self.create_thread(title, content, 2)

    async def send_markdown(
            self,
            title: str,
            content: str,
    ):
        """|coro|
        使用给定的内容向频道发送Markdown帖子。

        Parameters
        -----------
        title: :class:`str`
            帖子标题。
        content: :class:`str`
            帖子内容。
        """
        await self.create_thread(title, content, 3)

    async def send_json(
            self,
            title: str,
            content: str,
    ):
        """|coro|
        使用给定的内容向频道发送JSON帖子。

        Parameters
        -----------
        title: :class:`str`
            帖子标题。
        content: :class:`str`
            帖子内容。
        """
        await self.create_thread(title, content, 4)


class CategoryChannel(abc.GuildChannel, Hashable):
    """表示 QQ 频道类别。
    这些对于将频道分组到逻辑区间很有用。

    .. container:: operations

        .. describe:: x == y

            检查两个频道是否相等。

        .. describe:: x != y

            检查两个频道是否不相等。

        .. describe:: hash(x)

            返回类别的哈希值。

        .. describe:: str(x)

            返回类别的名称。

    Attributes
    -----------
    name: :class:`str`
        类别名称。
    guild: :class:`Guild`
        类别所属的公会。
    id: :class:`int`
        类别频道 ID。
    private_type: :class:`int`
        子频道私密类型。

        +----+---------------------+
        | 值 | 含义                |
        +----+---------------------+
        | 0  | 公开频道            |
        +----+---------------------+
        | 1  | 群主管理员可见      |
        +----+---------------------+
        | 2  | 群主管理员+指定成员 |
        +----+---------------------+

    position: :class:`int`
        在类别列表中的位置。 这是一个从 0 开始的数字。例如顶部频道是位置 0。
    """

    __slots__ = ('name', 'id', 'guild', '_state', 'position', 'category_id', 'private_type')

    def __init__(self, *, state: ConnectionState, guild: Guild, data: CategoryChannelPayload):
        self._state: ConnectionState = state
        self.id: int = int(data['id'])
        self._update(guild, data)

    def __repr__(self) -> str:
        return f'<CategoryChannel id={self.id} name={self.name!r} position={self.position}>'

    def _update(self, guild: Guild, data: CategoryChannelPayload) -> None:
        self.guild: Guild = guild
        self.name: str = data['name']
        self.category_id: Optional[int] = int(data.get('parent_id', 0))
        self.position: int = data['position']
        self.private_type: int = data.get('private_type')

    @property
    def _sorting_bucket(self) -> int:
        return ChannelType.category.value

    @property
    def type(self) -> ChannelType:
        """:class:`ChannelType`: 频道的 QQ 类型。"""
        return ChannelType.category

    @property
    def channels(self) -> List[GuildChannelType]:
        def comparator(channel):
            return not isinstance(channel, TextChannel), channel.position

        ret = [c for c in self.guild.channels if c.category_id == self.id]
        ret.sort(key=comparator)
        return ret

    @property
    def text_channels(self) -> List[TextChannel]:
        ret = [c for c in self.guild.channels if c.category_id == self.id and isinstance(c, TextChannel)]
        ret.sort(key=lambda c: (c.position, c.id))
        return ret

    @property
    def voice_channels(self) -> List[VoiceChannel]:
        ret = [c for c in self.guild.channels if c.category_id == self.id and isinstance(c, VoiceChannel)]
        ret.sort(key=lambda c: (c.position, c.id))
        return ret

    @property
    def live_channels(self) -> List[LiveChannel]:
        ret = [c for c in self.guild.channels if c.category_id == self.id and isinstance(c, LiveChannel)]
        ret.sort(key=lambda c: (c.position, c.id))
        return ret

    @property
    def app_channels(self) -> List[AppChannel]:
        ret = [c for c in self.guild.channels if c.category_id == self.id and isinstance(c, AppChannel)]
        ret.sort(key=lambda c: (c.position, c.id))
        return ret

    @property
    def thread_channels(self) -> List[ThreadChannel]:
        ret = [c for c in self.guild.channels if c.category_id == self.id and isinstance(c, ThreadChannel)]
        ret.sort(key=lambda c: (c.position, c.id))
        return ret


class PartialMessageable(abc.Messageable, Hashable):
    def __init__(self, state: ConnectionState, id: int, type: Optional[ChannelType] = None, direct: bool = False):
        self._state: ConnectionState = state
        self._channel: Object = Object(id=id)
        self.id: int = id
        self.type: Optional[ChannelType] = type
        self.direct = direct

    async def _get_channel(self) -> Tuple[Object, bool]:
        return self._channel, self.direct

    def get_partial_message(self, message_id: int, /) -> PartialMessage:
        from .message import PartialMessage

        return PartialMessage(channel=self, id=message_id)


def _guild_channel_factory(channel_type: int):
    value = try_enum(ChannelType, channel_type)
    if value is ChannelType.text:
        return TextChannel, value
    elif value is ChannelType.voice:
        return VoiceChannel, value
    elif value is ChannelType.category:
        return CategoryChannel, value
    elif value is ChannelType.live:
        return LiveChannel, value
    elif value is ChannelType.app:
        return AppChannel, value
    elif value is ChannelType.thread:
        return ThreadChannel, value
    else:
        return None, value


def _channel_factory(channel_type: int):
    cls, value = _guild_channel_factory(channel_type)
    return cls, value


DMC = TypeVar('DMC', bound='DMChannel')


class DMChannel(abc.Messageable, Hashable):
    """代表一个QQ私信子频道。
    
    .. container:: operations
        .. describe:: x == y
            检查两个子频道是否相等。
        .. describe:: x != y
            检查两个子频道是否不相等。
        .. describe:: hash(x)
            返回子频道的哈希值。
        .. describe:: str(x)
            返回子频道的字符串表示形式

    Attributes
    ----------
    recipient: Optional[:class:`User`]
        在私信子频道中参与的用户。如果通过网关接收此子频道，则接收者信息可能并不总是可用。
    me: :class:`ClientUser`
        代表你自己的用户。
    id: :class:`int`
        私信会话关联的子频道 id
    channel_id: :class:`int`
        垃圾资讯，无视即可，官方不知道在搞什么jb东西
    """

    __slots__ = ('id', 'recipient', 'me', 'channel_id', '_state')

    def __init__(self, *, me: ClientUser, state: ConnectionState, data: DMChannelPayload, recipients: User):
        self._state: ConnectionState = state
        self.recipient: Optional[User] = recipients
        self.me: ClientUser = me
        self.id: int = int(data['guild_id'])
        self.channel_id: int = int(data['channel_id'])

    async def _get_channel(self):
        return self, True

    def __str__(self) -> str:
        if self.recipient:
            return f'与 {self.recipient} 的直接消息'
        return '与未知用户的直接消息'

    def __repr__(self) -> str:
        return f'<DMChannel id={self.id} guild_id={self.channel_id} recipient={self.recipient!r}>'

    @classmethod
    def _from_message(cls: Type[DMC], state: ConnectionState, channel_id: int, guild_id: int) -> DMC:
        self: DMC = cls.__new__(cls)
        self._state = state
        self.channel_id = channel_id
        self.id = guild_id
        self.recipient = None
        # state.user won't be None here
        self.me = state.user  # type: ignore
        return self

    @property
    def type(self) -> ChannelType:
        """:class:`ChannelType`: 频道的 QQ 类型。"""
        return ChannelType.private

    def get_partial_message(self, message_id: int, /) -> PartialMessage:
        """从消息 ID 创建一个 :class:`PartialMessage` 。
        如果您想处理消息并且只拥有其 ID 而不进行不必要的 API 调用，这将非常有用。

        Parameters
        ------------
        message_id: :class:`int`
            要为其创建部分消息的消息 ID。

        Returns
        ---------
        :class:`PartialMessage`
            部分消息。
        """

        from .message import PartialMessage

        return PartialMessage(channel=self, id=message_id)



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import asyncio
from datetime import datetime
from typing import (
    overload, Optional, Union, List, TYPE_CHECKING, TypeVar, Dict, Any, runtime_checkable, Protocol, Tuple
)

from .enum import ChannelType
from .error import InvalidArgument
from .http import handle_message_parameters
from .iterators import HistoryIterator
from .utils import MISSING

if TYPE_CHECKING:
    from .user import ClientUser
    from .embeds import Ark, Embed, Markdown
    from .member import Member
    from .channel import CategoryChannel, TextChannel, PartialMessageable, DMChannel
    from .guild import Guild
    from .state import ConnectionState
    from .message import Message, MessageReference, PartialMessage
    from .types.channel import (
        Channel as ChannelPayload,
    )
    from .file import File

    PartialMessageableChannel = Union[TextChannel, PartialMessageable, DMChannel]
    MessageableChannel = Union[PartialMessageableChannel]

__all__ = (
    'Messageable',
    'GuildChannel',
    'BaseAudioControl'
)


class _Undefined:
    def __repr__(self) -> str:
        return 'see-below'


_undefined: Any = _Undefined()


@runtime_checkable
class PrivateChannel(Protocol):
    """详细说明私人QQ频道上的常见操作的抽象类。

    以下实现了这个 ABC：

    - :class:`~qq.DMChannel`

    Attributes
    -----------
    me: :class:`~discord.ClientUser`
        代表你自己的用户。
    """

    __slots__ = ()

    me: ClientUser
    id: int


class Messageable:
    """一个 记录了可以发送消息的模型上的常见操作的ABC。

    以下实现了这个 ABC：

    - :class:`~qq.TextChannel`
    - :class:`~qq.User`
    - :class:`~qq.Member`
    - :class:`~qq.ext.commands.Context`

    """

    __slots__ = ()
    _state: ConnectionState

    async def _get_channel(self) -> Tuple[MessageableChannel, bool]:
        raise NotImplementedError

    @overload
    async def send(
            self,
            content: Optional[str] = ...,
            *,
            image: str = ...,
            msg_id: Union[Message, MessageReference, PartialMessage, str] = ...,
            reference: Union[Message, MessageReference, PartialMessage] = ...,
            mention_author: Member = ...,
            ark: Ark = ...,
            embed: Embed = ...,
            markdown: Markdown = ...,
            file: File = ...,
            delete_after: float = ...,
    ) -> Union[Message, str]:
        ...

    async def send(
            self,
            content=None,
            *,
            image=None,
            msg_id='MESSAGE_CREATE',
            reference=None,
            mention_author=None,
            ark=None,
            embed=None,
            markdown=None,
            file=None,
            delete_after=None,
            replace_url=True,
    ):
        """|coro|
        使用给定的内容向目的地发送消息。
        content 必须是可以通过 ``str(content)`` 转换为字符串的类型。
        如果不填入 ``reference`` 将会被腾讯视为主动消息。
        如果是主动信息，输出将会是 ``audit_id``。

        Parameters
        ------------
        content: Optional[:class:`str`]
            发送的信息内容。
        image: :class:`str`
            要发送的图片链接
        ark: Optional[:class:`qq.Ark`]
            要发送的 Ark 类
        embed: Optional[:class:`qq.Embed`]
            要发送的 Embed 类
        markdown: Optional[:class:`qq.Markdown`]
            要发送的 Markdown 类
        msg_id: Union[:class:`~qq.Message`, :class:`~qq.MessageReference`, :class:`~qq.PartialMessage`]
            被动消息使用的消息

            .. note::

                如果不使用 ``msg_id`` ，系统将判断为主动消息，主动消息默认每天往每个频道可推送的消息数是 20 条，超过会被限制。

            .. note::

                目前官方放宽了被动消息，无视即可。

        reference: Union[:class:`~qq.Message`, :class:`~qq.MessageReference`, :class:`~qq.PartialMessage`]
            对你正在回复的 :class:`~qq.Message` 的引用，可以使用 :meth:`~qq.Message.to_reference` 创建或直接作为 :class:`~qq.Message` 传递。
        mention_author: Optional[:class:`qq.Member`]
            如果设置了，将会在消息前面提及该用户。
        file: Optional[:class:`qq.File`]
            本地上传的图片文件。
        delete_after: Optional[:class:`float`]
            如果设置了，则等待该秒数之后自动撤回消息。如果删除失败，则它会被静默忽略。
        replace_url: Optional[:class:`bool`]
            是否添加转义字符来绕过腾讯的链接检测。默认为 ``True`` 。

        Raises
        --------
        ~qq.HTTPException
            发送信息失败。
        ~qq.Forbidden
            你没有发送消息的适当权限。
        ~qq.InvalidArgument
            ``reference`` 不是 :class:`~qq.Message` 、
            :class:`~qq.MessageReference` 或 :class:`~qq.PartialMessage` 。

        Returns
        ---------
        Union[:class:`~qq.Message`, :class:`~qq.MessageAudit`]
            发送的消息。
        """
        from .channel import TextChannel
        channel, direct = await self._get_channel()
        state = self._state
        content = str(content) if content is not None else None

        if mention_author is not None:
            content = mention_author.mention + content

        # if msg_id is not None:
        #     try:
        #         msg_id = msg_id.to_message_reference_dict()
        #     except AttributeError:
        #         raise InvalidArgument(
        #             'msg_id 参数必须是 Message、 MessageReference 或 PartialMessage') from None

        if reference is not None:
            try:
                reference = reference.to_message_reference_dict()
            except AttributeError:
                raise InvalidArgument('reference 参数必须是 Message、 MessageReference 或 PartialMessage') from None

        with handle_message_parameters(
                content=content,
                direct=direct,
                msg_id=msg_id,
                file=file if file is not None else MISSING,
                image=image if image is not None else MISSING,
                embed=embed if embed is not None else MISSING,
                ark=ark if ark is not None else MISSING,
                markdown=markdown if markdown is not None else MISSING,
                message_reference=reference if reference is not None else MISSING,
                replace_url=replace_url,
        ) as params:
            data = await state.http.send_message(channel.id, params=params)

        # if msg_id is None:
        #     return data['data']['message_audit']['audit_id']

        ret = state.create_message(channel=channel, data=data, direct=direct)
        state.dispatch('message', ret)
        state._messages.append(ret)
        if channel and channel.__class__ in (TextChannel,):
            channel.last_message_id = ret.id
        if delete_after is not None:
            await ret.delete(delay=delete_after)  # type: ignore
        return ret

    async def fetch_message(self, id: str, /) -> Message:
        """|coro|
        从目的地检索单个 :class:`~qq.Message`。

        Parameters
        ------------
        id: :class:`str`
            要查找的消息 ID。

        Raises
        --------
        ~qq.NotFound
            未找到指定的消息。
        ~qq.Forbidden
            你没有获取消息所需的权限。
        ~qq.HTTPException
            检索消息失败。

        Returns
        --------
        :class:`~qq.Message`
            消息要求。
        """
        id = id
        channel, direct = await self._get_channel()
        data = await self._state.http.get_message(channel.id, id)
        return self._state.create_message(channel=channel, data=data['message'], direct=direct)

    def history(
            self,
            *,
            limit: Optional[int] = 100,
            before: Optional[datetime] = None,
            after: Optional[datetime] = None,
            around: Optional[datetime] = None,
            oldest_first: Optional[bool] = None,
    ) -> HistoryIterator:
        """返回允许接收目标消息历史记录的 :class:`~qq.AsyncIterator`。

        Examples
        ---------

        用途 ::

            counter = 0
            async for message in channel.history(limit=200):
                if message.author == client.user:
                    counter += 1

        扁平化为列表: ::

            messages = await channel.history(limit=123).flatten()
            # messages is now a list of Message...

        所有参数均为可选参数。

        Parameters
        -----------
        limit: Optional[:class:`int`]
            要检索的消息数。如果为 ``None`` ，则检索频道中的每条消息。但是请注意，这会使其运行缓慢。
        before: Optional[:class:`datetime.datetime`]
            检索此日期或消息之前的消息。如果提供了日期时间，建议使用 UTC 感知日期时间。如果 datetime 是本地的，则假定它是本地时间。
        after: Optional[:class:`datetime.datetime`]
            在此日期或消息之后检索消息。如果提供了日期时间，建议使用 UTC 感知日期时间。如果 datetime 是本地的，则假定它是本地时间。
        around: Optional[:class:`datetime.datetime`]
            检索围绕此日期或消息的消息。如果提供了日期时间，建议使用 UTC 感知日期时间。如果 datetime 是本地的，则假定它是本地时间。
            使用此参数时，最大限制为 101。
            请注意，如果限制为偶数，则最多返回 limit + 1 条消息。
        oldest_first: Optional[:class:`bool`]
            如果设置为 ``True``，以最旧->最新的顺序返回消息。如果指定了 ``after`` ，则默认为 ``True`` ，否则为 ``False`` 。

        Raises
        ------
        ~qq.Forbidden
            你无权获取频道消息历史记录。
        ~qq.HTTPException
            获取消息历史记录的请求失败。

        Yields
        -------
        :class:`~qq.Message`
            已解析消息数据的消息。
        """
        return HistoryIterator(self, limit=limit, before=before, after=after, around=around, oldest_first=oldest_first)


GCH = TypeVar('GCH', bound='GuildChannel')


class GuildChannel:
    """详细介绍 QQ 子频道上常见操作的 ABC。

    以下实现了这个 ABC：

    - :class:`~qq.TextChannel`
    - :class:`~qq.VoiceChannel`
    - :class:`~qq.CategoryChannel`
    - :class:`~qq.ThreadChannel`
    - :class:`~qq.LiveChannel`
    - :class:`~qq.AppChannel`

    Attributes
    -----------
    name: :class:`str`
        子频道名称。
    guild: :class:`~qq.Guild`
        子频道所属的频道。
    position: :class:`int`
        在频道列表中的位置。这是一个从 0 开始的数字。例如顶部子是位置 0。
    """

    __slots__ = ()

    id: int
    name: str
    guild: Guild
    type: ChannelType
    position: int
    category_id: Optional[int]
    _state: ConnectionState

    if TYPE_CHECKING:
        def __init__(self, *, state: ConnectionState, guild: Guild, data: Dict[str, Any]):
            ...

    def __str__(self) -> str:
        return self.name

    @property
    def _sorting_bucket(self) -> int:
        raise NotImplementedError

    def _update(self, guild: Guild, data: Dict[str, Any]) -> None:
        raise NotImplementedError

    async def _move(
            self,
            position: int,
            parent_id: Optional[Any] = None,
            *,
            reason: Optional[str],
    ) -> None:
        if position < 0:
            raise InvalidArgument('频道位置不能小于 0。')

        http = self._state.http
        bucket = self._sorting_bucket
        channels: List[GuildChannel] = [c for c in self.guild.channels if c._sorting_bucket == bucket]

        channels.sort(key=lambda c: c.position)

        try:
            # remove ourselves from the channel list
            channels.remove(self)
        except ValueError:
            # not there somehow lol
            return
        else:
            index = next((i for i, c in enumerate(channels) if c.position >= position), len(channels))
            # add ourselves at our designated position
            channels.insert(index, self)

        payload = []
        for index, c in enumerate(channels):
            d: Dict[str, Any] = {'id': c.id, 'position': index}
            if parent_id is not _undefined and c.id == self.id:
                d.update(parent_id=parent_id)
            payload.append(d)

        await asyncio.gather(*http.bulk_channel_update(self.guild.id, payload, reason=reason))

    async def _edit(self, options: Dict[str, Any], reason: Optional[str]) -> Optional[ChannelPayload]:
        try:
            parent = options.pop('category')
        except KeyError:
            parent_id = _undefined
        else:
            parent_id = parent and parent.id

        try:
            position = options.pop('position')
        except KeyError:
            if parent_id is not _undefined:
                options['parent_id'] = parent_id
        else:
            await self._move(position, parent_id=parent_id, reason=reason)

        try:
            ch_type = options['type']
        except KeyError:
            pass
        else:
            if not isinstance(ch_type, ChannelType):
                raise InvalidArgument('type 字段必须是 ChannelType 类型')
            options['type'] = ch_type.value

        if options:
            return await self._state.http.edit_channel(self.id, reason=reason, **options)

    @property
    def mention(self) -> str:
        """:class:`str`: 允许你提及频道的字符串。"""
        return f'<#{self.id}>'

    @property
    def category(self) -> Optional[CategoryChannel]:
        """Optional[:class:`~qq.CategoryChannel`]: 此频道所属的类别。如果没有类别，则为 ``None``。
        """
        return self.guild.get_channel(self.category_id)  # type: ignore

    async def delete(self, *, reason: Optional[str] = None) -> None:
        """|coro|
        删除频道。
        """
        await self._state.http.delete_channel(self.id, reason=reason)

    async def _clone_impl(
            self: GCH,
            base_attrs: Dict[str, Any],
            *,
            name: Optional[str] = None,
            reason: Optional[str] = None,
    ) -> GCH:
        base_attrs['parent_id'] = self.category_id
        base_attrs['name'] = name or self.name
        guild_id = self.guild.id
        cls = self.__class__
        data = await self._state.http.create_channel(guild_id, self.type.value, reason=reason, **base_attrs)
        obj = cls(state=self._state, guild=self.guild, data=data)

        # temporarily add it to the cache
        self.guild._channels[obj.id] = obj  # type: ignore
        return obj

    async def clone(self: GCH, *, name: Optional[str] = None, reason: Optional[str] = None) -> GCH:
        raise NotImplementedError

    @overload
    async def move(
            self,
            *,
            beginning: bool,
            offset: int = MISSING,
            category: Optional[int] = MISSING,
            sync_permissions: bool = MISSING,
            reason: Optional[str] = MISSING,
    ) -> None:
        ...

    @overload
    async def move(
            self,
            *,
            end: bool,
            offset: int = MISSING,
            category: Optional[int] = MISSING,
            sync_permissions: bool = MISSING,
            reason: str = MISSING,
    ) -> None:
        ...

    @overload
    async def move(
            self,
            *,
            before: int,
            offset: int = MISSING,
            category: Optional[int] = MISSING,
            sync_permissions: bool = MISSING,
            reason: str = MISSING,
    ) -> None:
        ...

    @overload
    async def move(
            self,
            *,
            after: int,
            offset: int = MISSING,
            category: Optional[int] = MISSING,
            sync_permissions: bool = MISSING,
            reason: str = MISSING,
    ) -> None:
        ...

    async def move(self, **kwargs) -> None:
        """|coro|
        帮助你相对于其他频道移动频道。
        如果需要精确的位置移动，则应使用 ``edit`` 代替。

        Parameters
        ------------
        beginning: :class:`bool`
            是否将频道移动到频道列表（或类别，如果给定）的开头。这与 ``end`` 、 ``before`` 和 ``after`` 是互斥的。
        end: :class:`bool`
            是否将频道移动到频道列表（或类别，如果给定）的末尾。这与 ``beginning`` 、 ``before`` 和 ``after`` 是互斥的。
        before: :class:`GuildChannel`
            应该在我们当前频道之前的频道。这与 ``beginning`` 、 ``end`` 和` `after`` 是互斥的。
        after: :class:`GuildChannel`
            应该在我们当前频道之后的频道。这与 ``beginning`` 、 ``end`` 和 ``before`` 是互斥的。
        offset: :class:`int`
            偏移移动的通道数。
            例如，带有 ``beginning=True`` 的 ``2`` 偏移量会在开始后移动 2。
            正数将其移至下方，而负数将其移至上方。请注意，这个数字是相对的，并且是在 ``beginning`` 、 ``end`` 、 ``before`` 和 ``after`` 参数之后计算的。
        category: Optional[:class:`GuildChannel`]
            将此频道移动到的类别。如果给出 ``None``，则将其移出类别。如果移动类别频道，则忽略此参数。

        Raises
        -------
        InvalidArgument
            给出了无效的位置或传递了错误的参数组合。
        Forbidden
            你无权移动频道。
        HTTPException
            移动频道失败。
        """

        if not kwargs:
            return

        beginning, end = kwargs.get('beginning'), kwargs.get('end')
        before, after = kwargs.get('before'), kwargs.get('after')
        offset = kwargs.get('offset', 0)
        if sum(bool(a) for a in (beginning, end, before, after)) > 1:
            raise InvalidArgument('只能使用 [before, after, end, begin] 之一。')

        bucket = self._sorting_bucket
        parent_id = kwargs.get('category', MISSING)
        # fmt: off
        channels: List[GuildChannel]
        if parent_id not in (MISSING, None):
            parent_id = parent_id.id
            channels = [ch for ch in self.guild.channels
                        if ch._sorting_bucket == bucket and ch.category_id == parent_id]
        else:
            channels = [ch for ch in self.guild.channels
                        if ch._sorting_bucket == bucket and ch.category_id == self.category_id]
        # fmt: on

        channels.sort(key=lambda c: (c.position, c.id))

        try:
            # Try to remove ourselves from the channel list
            channels.remove(self)
        except ValueError:
            # If we're not there then it's probably due to not being in the category
            pass

        index = None
        if beginning:
            index = 0
        elif end:
            index = len(channels)
        elif before:
            index = next((i for i, c in enumerate(channels) if c.id == before.id), None)
        elif after:
            index = next((i + 1 for i, c in enumerate(channels) if c.id == after.id), None)

        if index is None:
            raise InvalidArgument('无法解析适当的移动位置')

        channels.insert(max((index + offset), 0), self)
        payload = []
        for index, channel in enumerate(channels):
            d = {'id': channel.id, 'position': index}
            if parent_id is not MISSING and channel.id == self.id:
                d.update(parent_id=parent_id)
            payload.append(d)

        await asyncio.gather(*self._state.http.bulk_channel_update(self.guild.id, payload))


class BaseAudioControl:
    __slots__ = (
        '_audio_url',
        '_text',
    )

    def __str__(self) -> str:
        return self._text

    def __repr__(self) -> str:
        attrs = [
            ('audio_url', self.audio_url),
            ('text', self._text),
        ]
        joined = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {joined}>'

    def set_url(self, url: str):
        """
        设定要开始的音频地址。

        Parameters
        ------------
        url: :class:`str`
            设定的音频地址
        """
        self._audio_url = url

    def set_text(self, text: str):
        """
        设定要开始的音频状态文本。

        Parameters
        ------------
        text: :class:`str`
            设定的音频状态文本
        """
        self._text = text

    @property
    def audio_url(self):
        return self._audio_url

    @property
    def text(self):
        return self._text



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Optional

from .member import Member
from .types.schedule import Schedule as SchedulePayload

__all__ = ('Schedule',)

if TYPE_CHECKING:
    from .abc import GuildChannel
    from .state import ConnectionState
    from .guild import Guild


class Schedule:
    """代表一个 :class:`AppChannel` 的日历。。

    Attributes
    ----------
    id: :class:`str`
        一个 datetime 对象，它指定成员加入频道的日期和时间。
        如果成员离开并重新加入频道，这将是最新的日期。在某些情况下，这可以是 ``None`` 。
    name: :class:`str`
        成员所属的频道。
    guild: :class:`Guild`
        日程所在的频道
    channel: :class:`GuildChannel`
        日程所在的频道
    description: Optional[:class:`str`]
        用户的频道特定昵称。
    start: :class:`datatime.datetime`
        日程开始的时间。
    end: start: :class:`datatime.datetime`
        日程结束的时间。
    creator: :class:`Member`
        日程创建者。
    jump_channel: :class:`GuildChannel`
        日程跳转频道。
    remind_type: :class:`str`
        日程提醒类型。

        +-----------+--------------+
        | 提醒类型 id	  | 描述           |
        +===========+==============+
        | 0         | 不提醒          |
        | 1         | 开始时提醒        |
        | 2         | 开始前 5 分钟提醒   |
        | 3         | 开始前 15 分钟提醒  |
        | 4         | 开始前 30 分钟提醒  |
        | 5         | 开始前 60 分钟提醒  |
        +-----------+--------------+

    """

    __slots__ = (
        '_state',
        'id',
        'name',
        'description',
        'start',
        'end',
        'creator',
        'jump_channel',
        'remind_type',
        'channel',
        'guild'
    )

    if TYPE_CHECKING:
        name: str
        id: int
        description: Optional[str]

    def __init__(self, data: SchedulePayload, state: ConnectionState, guild: Guild, channel: GuildChannel):
        self.id = int(data['id'])
        self.name = data['name']
        self.description = data.get('description', None)
        self.start = datetime.datetime.fromtimestamp(int(data['start_timestamp']) / 1000)
        self.end = datetime.datetime.fromtimestamp(int(data['end_timestamp']) / 1000)
        self.creator = Member(data=data['creator'], guild=guild, state=state)
        self.jump_channel = guild.get_channel(int(data['jump_channel_id']))
        self.remind_type = data['remind_type']
        self._state = state
        self.channel = channel
        self.guild = guild

    @classmethod
    def from_id(cls, channel: GuildChannel, schedule_id: int):
        """
        创建一个不完全的 :class:`Schedule` ，主要用语 :meth:`Schedule.delete`
        """
        self = cls.__new__(cls)
        self.channel = channel
        self.id = schedule_id
        return self

    async def delete(self, reason: Optional[str] = None):
        """|coro|
        删除这个日历。
        """
        await self._state.http.remove_schedule(self.channel.id, self.id, reason=reason)



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from .channel import TextChannel

if TYPE_CHECKING:
    from typing import Coroutine
    from .state import ConnectionState
    from .guild import Guild
    from .types.permission import (
        Permission as PermissionPayload,
    )

__all__ = ('Permission',)


class Permission:
    """
    代表机器人 API 接口权限

    Attributes
    ----------
    path: :class:`str`
        API 接口名，例如 ``/guilds/{guild_id}/members/{user_id}``
    method: :class:`str`
        请求方法，例如 ``GET``
    desc: :class:`str`
        API 接口名称，例如 ``获取频道信``
    """
    __slots__ = (
        'path',
        'method',
        'desc',
        '_auth_status',
        '_state',
        '_guild'
    )

    def __init__(self, data: PermissionPayload, state: ConnectionState, guild: Guild):
        self.path = data.get('path')
        self.method = data.get('method')
        self.desc = data.get('desc')
        self._auth_status = data.get('auth_status')
        self._state = state
        self._guild = guild

    def __repr__(self):
        return f"<Permission desc={self.desc} path={self.method} {self.path}>"

    @property
    def enabled(self):
        """:class:`bool`: 是否已经启用。"""
        return self._auth_status == 1

    @property
    def disabled(self):
        """:class:`bool`: 是否已经禁用。"""
        return self._auth_status != 1

    def demand(self, channel: TextChannel, desc=None) -> Coroutine:
        return self._state.http.demand_permission(
            guild_id=self._guild.id,
            channel_id=channel.id,
            desc=desc or self.desc,
            path=self.path,
            method=self.method
        )



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING

from .guild import Guild

if TYPE_CHECKING:
    from .types.interaction import Interaction as InteractionPayload
    from .state import ConnectionState


class Interaction:
    """代表一个互动。

    请注意你需要在结束处理，或者判断权限结束后通过 :meth:`Interaction.success` 等函数回应互动，不然会导致超时。

    你也可以使用 async with 表达式在离开作用域时，如果未曾响应则自动调用 :meth:`Interaction.success` 回复成功。

    Attributes
    ----------
    application_id: :class:`str`
        应用名称
    id: :class:`int`
        互动 ID
    guild: :class:`Guild`
        互动所发生的频道
    channel: :class:`Channel`
        互动所发生的子频道
    message: :class:`Message`
        互动所发生的消息
    author: :class:`Member`
        互动造成的用户
    chat_type: :class:`int`
        聊天类型？
    button_data: :class:`str`
        按钮名称/数据
    button_id: :class:`str`
        按钮 ID
    version: Optional[:class:`str`]
        互动版本
    """
    __slots__ = (
        "application_id",
        "chat_type",
        "button_data",
        "button_id",
        "_user_id",
        "_guild_id",
        "_channel_id",
        "_message_id",
        "guild",
        "channel",
        "message",
        "author",
        "type",
        "version",
        "id",
        "_state",
        "_responded"
    )

    def __init__(self, state: ConnectionState, data: InteractionPayload):
        self._state = state
        self.application_id = data["application_id"]
        self.chat_type = data["chat_type"]
        self._channel_id = int(data["channel_id"])
        self._guild_id = int(data["guild_id"])
        if "data" in data and "resolved" in data["data"]:
            rs = data["data"]["resolved"]
            self.button_data = rs["button_data"]
            self.button_id = rs["button_id"]
            self._user_id = int(rs["user_id"])
            self._message_id = rs["message_id"]
        self.author = None
        self.guild = None
        self.channel = None
        self.message = None
        self.type = data["type"]
        self.version = data["version"]
        self.id = data["id"]
        self._responded = False

    async def upgrade(self):
        self.guild = self._state._get_guild(self._guild_id)
        if self.guild is None:
            data = await self._state.http.get_guild(self._guild_id)
            self.guild = Guild(data=data, state=self._state)
        self.channel = self.guild.get_channel(self._channel_id)
        if self.channel is None:
            await self.guild.fetch_channel(self._channel_id)
        self.author = self.guild.get_member(self._user_id)
        if self.author is None:
            self.author = await self.guild.fetch_member(self._user_id)
        self.message = await self.channel.fetch_message(self._message_id)

    async def __aenter__(self):
        return self

    async def __aexit__(self):
        if not self._responded:
            await self.success()

    async def _ack(self, code: int):
        if self._responded:
            return
        self._responded = True
        await self._state.http.ack_interaction(self.id, code)

    async def success(self):
        """|coro|

        回复操作成功。
        """

        await self._ack(0)

    async def failed(self):
        """|coro|

        回复操作失败。
        """

        await self._ack(1)

    async def too_frequent(self):
        """|coro|

        回复操作过于频繁。
        """

        await self._ack(2)

    async def duplicated(self):
        """|coro|

        回复操作为重复操作。
        """

        await self._ack(3)

    async def no_permission(self):
        """|coro|

        回复操作权限不足。
        """

        await self._ack(4)

    async def only_admin(self):
        """|coro|

        回复操作仅能被管理员调用。
        """

        await self._ack(5)



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import asyncio
import datetime
import logging
import sys
import weakref
from types import TracebackType
from typing import (
    ClassVar, Any, Optional, Dict, Union, TypeVar, Type,
    Coroutine, List, Tuple, TYPE_CHECKING, NamedTuple
)
from urllib.parse import quote as _uriquote

import aiohttp

from . import __version__, utils
from .embeds import Ark, Embed, Markdown
from .error import HTTPException, Forbidden, NotFound, QQServerError, LoginFailure, GatewayNotFound
from .gateway import QQClientWebSocketResponse
from .types import user, guild, message, channel, member
from .types.forum import (
    ThreadListResponse,
    ThreadInfo as ThreadPayload,
)
from .types.message import Message
from .types.permission import (
    Permission as PermissionPayload,
    PermissionDemand as PermissionDemandPayload
)
from .types.role import (
    WrappedRole as WrappedRolePayload,
    Role as RolePayload
)
from .types.schedule import Schedule as SchedulePayload
from .utils import MISSING

if TYPE_CHECKING:
    from .message import MessageReference
    from .file import File

T = TypeVar('T')
BE = TypeVar('BE', bound=BaseException)
MU = TypeVar('MU', bound='MaybeUnlock')
Response = Coroutine[Any, Any, T]
_log = logging.getLogger(__name__)

__all__ = ('Route', 'HTTPClient')


class Route:
    BASE: ClassVar[str] = 'https://api.sgroup.qq.com'

    def __init__(self, method: str, path: str, **parameters: Any) -> None:
        self.path: str = path
        self.method: str = method
        url = self.BASE + self.path
        if parameters:
            url = url.format_map({k: _uriquote(v) if isinstance(v, str) else v for k, v in parameters.items()})
        self.url: str = url

        # major parameters:
        self.channel_id: Optional[str] = parameters.get('channel_id')
        self.guild_id: Optional[str] = parameters.get('guild_id')
        self.token: Optional[str] = parameters.get('token')

    @property
    def bucket(self) -> str:
        # the bucket is just method + path w/ major parameters
        return f'{self.channel_id}:{self.guild_id}:{self.path}'


async def json_or_text(response: aiohttp.ClientResponse) -> Union[Dict[str, Any], str]:
    text = await response.text(encoding='utf-8')
    try:
        if response.headers['content-type'] == 'application/json':
            return utils._from_json(text)
    except Exception:
        # Thanks Cloudflare
        pass

    return text


class MultipartParameters(NamedTuple):
    direct: bool
    payload: Optional[Dict[str, Any]]
    multipart: Optional[Dict[str, Any]]
    file: Optional[File]

    def __enter__(self) -> MultipartParameters:
        return self

    def __exit__(
            self,
            exc_type: Optional[Type[BE]],
            exc: Optional[BE],
            traceback: Optional[TracebackType],
    ) -> None:
        if self.file:
            self.file.close()


def handle_message_parameters(
        content: Optional[str] = MISSING,
        direct: bool = False,
        *,
        msg_id: Optional[str] = MISSING,
        file: File = MISSING,
        image: Optional[str] = MISSING,
        embed: Optional[Embed] = MISSING,
        ark: Optional[Ark] = MISSING,
        markdown: Optional[Markdown] = MISSING,
        message_reference: Optional[MessageReference] = MISSING,
        replace_url: bool = True,
) -> MultipartParameters:
    payload = {}
    if msg_id:
        payload['msg_id'] = msg_id

    if embed:
        payload['embed'] = embed.to_dict()

    if ark:
        payload['ark'] = ark.to_dict()

    if markdown:
        payload['markdown'] = markdown.to_dict()
        if markdown._keyboard is not None:
            payload['keyboard'] = markdown._keyboard.to_dict()

    if image:
        payload['image'] = image

    if content is not MISSING:
        if content is not None:
            payload['content'] = str(content)
        else:
            pass

    if replace_url and 'content' in payload:
        payload['content'] = payload['content'].replace(".", "\ufeff.")

    if message_reference is not MISSING:
        payload['message_reference'] = message_reference

    multipart = {}
    if file:
        for key, value in payload.items():
            multipart[key] = value
        payload = None
        multipart['file_image'] = file.fp

    return MultipartParameters(direct=direct, payload=payload, multipart=multipart, file=file)


class MaybeUnlock:
    def __init__(self, lock: asyncio.Lock) -> None:
        self.lock: asyncio.Lock = lock
        self._unlock: bool = True

    def __enter__(self: MU) -> MU:
        return self

    def defer(self) -> None:
        self._unlock = False

    def __exit__(
            self,
            exc_type: Optional[Type[BE]],
            exc: Optional[BE],
            traceback: Optional[TracebackType],
    ) -> None:
        if self._unlock:
            self.lock.release()


class HTTPClient:
    """Represents an HTTP client sending HTTP requests to the QQ API."""

    def __init__(
            self,
            connector: Optional[aiohttp.BaseConnector] = None,
            *,
            proxy: Optional[str] = None,
            proxy_auth: Optional[aiohttp.BasicAuth] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            unsync_clock: bool = True,
    ) -> None:
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop() if loop is None else loop
        self.connector = connector
        self.__session: aiohttp.ClientSession = MISSING  # filled in static_login
        self._locks: weakref.WeakValueDictionary = weakref.WeakValueDictionary()
        self._global_over: asyncio.Event = asyncio.Event()
        self._global_over.set()
        self.token: Optional[str] = None
        self.bot_token: bool = False
        self.proxy: Optional[str] = proxy
        self.proxy_auth: Optional[aiohttp.BasicAuth] = proxy_auth
        self.use_clock: bool = not unsync_clock

        user_agent = "QQBot (https://github.com/foxwhite25/qq.py {0}) Python/{1[0]}.{1[1]} aiohttp/{2}"
        self.user_agent: str = user_agent.format(__version__, sys.version_info, aiohttp.__version__)

    async def request(
            self,
            route: Route,
            *,
            file: Optional[File] = None,
            form: Optional[Dict[str, Any]] = None,
            **kwargs: Any,
    ) -> Any:
        bucket = route.bucket
        method = route.method
        url = route.url

        lock = self._locks.get(bucket)
        if lock is None:
            lock = asyncio.Lock()
            if bucket is not None:
                self._locks[bucket] = lock

        headers: Dict[str, str] = {
            'User-Agent': self.user_agent,
        }

        # Add token to header
        if self.token is not None:
            headers['Authorization'] = 'Bot ' + self.token

        # Checking if it's a JSON request
        if 'json' in kwargs:
            headers['Content-Type'] = 'application/json'
            kwargs['data'] = utils._to_json(kwargs.pop('json'))

        kwargs['headers'] = headers

        # Proxy support
        if self.proxy is not None:
            kwargs['proxy'] = self.proxy
        if self.proxy_auth is not None:
            kwargs['proxy_auth'] = self.proxy_auth

        if not self._global_over.is_set():
            # wait until the global lock is complete
            await self._global_over.wait()

        try:
            reason = kwargs.pop('reason')
        except KeyError:
            pass
        else:
            if reason:
                headers['X-Audit-Log-Reason'] = _uriquote(reason, safe='/ ')

        response: Optional[aiohttp.ClientResponse] = None
        data: Optional[Union[Dict[str, Any], str]] = None
        await lock.acquire()
        with MaybeUnlock(lock) as maybe_lock:
            for tries in range(5):
                if file:
                    file.reset(seek=tries)

                if form:
                    form_data = aiohttp.FormData()
                    for key, value in form.items():
                        form_data.add_field(key, value)
                    kwargs['data'] = form_data
                try:
                    async with self.__session.request(method, url, **kwargs) as response:
                        _log.debug(
                            '%s %s 与 %s 已返回 %s Trace ID: %s', method, url, kwargs.get('data'),
                            response.status,
                            response.headers['X-Tps-trace-ID'] if 'X-Tps-trace-ID' in response.headers else "Missing"
                        )

                        # even errors have text involved in them so this is safe to call
                        data = await json_or_text(response)

                        # the request was successful so just return the text/json
                        if 300 > response.status >= 200:
                            if response.status != 204:
                                _log.debug('%s %s 已收到 %s', method, url, data)
                            return data

                        # we've received a 500, 502, or 504, unconditional retry
                        # if response.status in (500, 502, 504):
                        #     await asyncio.sleep(1 + tries * 2)
                        #     continue

                        # the usual error cases
                        if response.status in [404, 403, 401]:
                            raise Forbidden(response, data, route=route)
                        elif response.status == 404:
                            raise NotFound(response, data, route=route)
                        elif response.status >= 500:
                            raise QQServerError(response, data, route=route)
                        else:
                            raise HTTPException(response, data, route=route)

                    # This is handling exceptions from the request
                except OSError as e:
                    # Connection reset by peer
                    if tries < 4 and e.errno in (54, 10054):
                        await asyncio.sleep(1 + tries * 2)
                        continue
                    raise
                except QQServerError as e:
                    if tries < 4 and e.code in (620006,):
                        await asyncio.sleep(1 + tries * 2)
                        continue
                    raise e

                if response is not None:
                    # We've run out of retries, raise.
                    if response.status >= 500:
                        raise QQServerError(response, data)

                    raise HTTPException(response, data)

                raise RuntimeError('HTTP 处理中无法访问的代码')

    async def static_login(self, token: str) -> user.User:
        # Necessary to get aiohttp to stop complaining about session creation
        self.__session = aiohttp.ClientSession(connector=self.connector, ws_response_class=QQClientWebSocketResponse)
        old_token = self.token
        self.token = token

        try:
            data = await self.request(Route('GET', '/users/@me'))
        except HTTPException as exc:
            self.token = old_token
            if exc.status == 401:
                raise LoginFailure('传递了不正确的令牌。') from exc
            raise

        return data

    def get_guilds(
            self,
            limit: int = 100,
            before: Optional[str] = None,
            after: Optional[str] = None,
    ) -> Response[List[guild.Guild]]:
        params: Dict[str, Any] = {
            'limit': limit,
        }

        if before:
            params['before'] = before
        if after:
            params['after'] = after

        return self.request(Route('GET', '/users/@me/guilds'), params=params)

    def get_guild(self, guild_id: int) -> Response[guild.Guild]:
        return self.request(Route('GET', '/guilds/{guild_id}', guild_id=guild_id))

    def get_guild_channels(self, guild_id: int) -> Response[guild.Guild]:
        return self.request(Route('GET', '/guilds/{guild_id}/channels', guild_id=guild_id))

    def get_message(self, channel_id: int, message_id: str) -> Response[Message]:
        r = Route('GET', '/channels/{channel_id}/messages/{message_id}', channel_id=channel_id, message_id=message_id)
        return self.request(r)

    # 身份组管理

    def get_roles(self, guild_id: int) -> Response[List[RolePayload]]:
        return self.request(Route('GET', '/guilds/{guild_id}/roles', guild_id=guild_id))

    def edit_role(
            self, guild_id: int, role_id: int, *, reason: Optional[str] = None, **fields: Any
    ) -> Response[RolePayload]:
        r = Route('PATCH', '/guilds/{guild_id}/roles/{role_id}', guild_id=guild_id, role_id=role_id)
        valid_keys = ('name', 'color', 'hoist')
        payload = {"info": {k: v for k, v in fields.items() if k in valid_keys}}
        return self.request(r, json=payload, reason=reason)

    def delete_role(self, guild_id: int, role_id: int, *, reason: Optional[str] = None) -> Response[None]:
        r = Route('DELETE', '/guilds/{guild_id}/roles/{role_id}', guild_id=guild_id, role_id=role_id)
        return self.request(r, reason=reason)

    def create_role(
            self, guild_id: int, *, reason: Optional[str] = None, **fields: Any
    ) -> Response[WrappedRolePayload]:
        r = Route('POST', '/guilds/{guild_id}/roles', guild_id=guild_id)
        return self.request(r, json=fields, reason=reason)

    def add_role(
            self, guild_id: int, user_id: int, role_id: int, channel_id: Optional[int], *, reason: Optional[str] = None
    ) -> Response[None]:
        r = Route(
            'PUT',
            '/guilds/{guild_id}/members/{user_id}/roles/{role_id}',
            guild_id=guild_id,
            user_id=user_id,
            role_id=role_id,
        )
        if channel_id:
            payload = {'channel': {'id': str(channel_id)}}
            return self.request(r, json=payload, reason=reason)
        return self.request(r, reason=reason)

    def remove_role(
            self, guild_id: int, user_id: int, role_id: int, channel_id: Optional[int], *, reason: Optional[str] = None
    ) -> Response[None]:
        r = Route(
            'DELETE',
            '/guilds/{guild_id}/members/{user_id}/roles/{role_id}',
            guild_id=guild_id,
            user_id=user_id,
            role_id=role_id,
        )
        if channel_id:
            payload = {'channel': {'id': str(channel_id)}}
            return self.request(r, json=payload, reason=reason)
        return self.request(r, reason=reason)

    def create_channel(
            self,
            guild_id: int,
            channel_type: channel.ChannelType,
            *,
            reason: Optional[str] = None,
            **options: Any,
    ) -> Response[channel.GuildChannel]:
        payload = {
            'type': channel_type,
        }

        valid_keys = (
            'name',
            'parent_id',
            'position',
            'private_type',
            'speak_permission',
            'application_id',
        )
        payload.update({k: str(v) for k, v in options.items() if k in valid_keys and v is not None})

        if 'private_user_ids' in options:
            payload['private_user_ids'] = [str(x) for x in options['private_user_ids']]

        return self.request(
            Route('POST', '/guilds/{guild_id}/channels', guild_id=guild_id),
            json=payload,
            reason=reason
        )

    def edit_channel(
            self,
            channel_id: int,
            *,
            reason: Optional[str] = None,
            **options: Any,
    ) -> Response[channel.Channel]:
        r = Route('PATCH', '/channels/{channel_id}', channel_id=channel_id)
        valid_keys = (
            'name',
            'parent_id',
            'position',
            'type',
            'private_type',
            'speak_permission',
        )
        payload = {k: v for k, v in options.items() if k in valid_keys}
        return self.request(r, reason=reason, json=payload)

    def bulk_channel_update(
            self,
            guild_id: int,
            datas: List[guild.ChannelPositionUpdate],
            *,
            reason: Optional[str] = None,
    ) -> List[Response[None]]:
        rsp = []
        for data in datas:
            valid_keys = (
                'name',
                'parent_id',
                'position',
                'type',
            )
            payload = {k: v for k, v in data.items() if k in valid_keys}
            r = Route('PATCH', '/channels/{channel_id}', channel_id=data.get('id'))
            rsp.append(self.request(r, reason=reason, json=payload))
        return rsp

    def delete_channel(
            self,
            channel_id: int,
            *,
            reason: Optional[str] = None,
    ) -> Response[None]:
        return self.request(Route('DELETE', '/channels/{channel_id}', channel_id=channel_id), reason=reason)

    async def get_from_cdn(self, url: str) -> bytes:
        async with self.__session.get(url) as resp:
            if resp.status == 200:
                return await resp.read()
            elif resp.status == 404:
                raise NotFound(resp, 'asset not found')
            elif resp.status == 403:
                raise Forbidden(resp, 'cannot retrieve asset')
            else:
                raise HTTPException(resp, 'failed to get asset')

    async def get_gateway(self, *, encoding: str = 'json', zlib: bool = True) -> str:
        try:
            data = await self.request(Route('GET', '/gateway'))
        except HTTPException as exc:
            raise GatewayNotFound() from exc
        return data['url']

    async def get_bot_gateway(self, *, encoding: str = 'json', zlib: bool = True) -> Tuple[int, str]:
        try:
            data = await self.request(Route('GET', '/gateway/bot'))
        except HTTPException as exc:
            raise GatewayNotFound() from exc

        if zlib:
            value = '{0}?encoding={1}&v=9&compress=zlib-stream'
        else:
            value = '{0}?encoding={1}&v=9'
        return data['shards'], value.format(data['url'], encoding)

    def recreate(self) -> None:
        if self.__session.closed:
            self.__session = aiohttp.ClientSession(
                connector=self.connector, ws_response_class=QQClientWebSocketResponse
            )

    async def close(self) -> None:
        if self.__session:
            await self.__session.close()

    async def ws_connect(self, url: str, *, compress: int = 0) -> Any:
        kwargs = {
            'proxy_auth': self.proxy_auth,
            'proxy': self.proxy,
            'max_msg_size': 0,
            'timeout': 30.0,
            'autoclose': False,
            'headers': {
                'User-Agent': self.user_agent,
            },
            'compress': compress,
        }

        return await self.__session.ws_connect(url, **kwargs)

    def start_private_message(self, user_id: int, guild_id: int) -> Response[channel.DMChannel]:
        payload = {
            'recipient_id': str(user_id),
            'source_guild_id': str(guild_id),
        }

        return self.request(Route('POST', '/users/@me/dms'), json=payload)

    def send_message(
            self,
            channel_id: int,
            *,
            params: MultipartParameters,
    ) -> Response[message.Message]:
        r = Route('POST', '/channels/{channel_id}/messages', channel_id=channel_id) if not params.direct else \
            Route('POST', '/dms/{guild_id}/messages', guild_id=channel_id)

        if params.file:
            return self.request(r, file=params.file, form=params.multipart)
        else:
            return self.request(r, json=params.payload)

    def get_members(
            self, guild_id: int, limit: int, after: Optional[int] = None
    ) -> Response[List[member.MemberWithUser]]:
        params: Dict[str, Any] = {
            'limit': limit,
        }
        if after:
            params['after'] = after

        r = Route('GET', '/guilds/{guild_id}/members', guild_id=guild_id)
        return self.request(r, params=params)

    def get_member(self, guild_id: int, member_id: int) -> Response[member.MemberWithUser]:
        return self.request(
            Route('GET', '/guilds/{guild_id}/members/{member_id}', guild_id=guild_id, member_id=member_id))

    def get_channel(self, channel_id: int) -> Response[channel.Channel]:
        r = Route('GET', '/channels/{channel_id}', channel_id=channel_id)
        return self.request(r)

    def kick(self, user_id: int, guild_id: int, add_blacklist: bool, reason: Optional[str] = None) -> Response[None]:
        r = Route('DELETE', '/guilds/{guild_id}/members/{user_id}', guild_id=guild_id, user_id=user_id)
        payload: Dict[str, Any] = {
            'add_blacklist': add_blacklist,
        }
        if reason:
            # thanks aiohttp
            r.url = f'{r.url}?reason={_uriquote(reason)}'

        return self.request(r, json=payload)

    def create_schedule(
            self,
            channel_id: int,
            name: str,
            start_timestamp: Union[datetime.datetime, float],
            end_timestamp: Union[datetime.datetime, float],
            jump_channel_id: int,
            remind_type: str,
            description: Optional[str],
            reason: Optional[str] = None
    ) -> Response[SchedulePayload]:
        payload: Dict[str, Any] = {
            "schedule": {
                "name": name,
                "start_timestamp": str(int(start_timestamp * 1000)) if isinstance(start_timestamp, float)
                else str(int(start_timestamp.timestamp() * 1000)),
                "end_timestamp": str(int(end_timestamp * 1000)) if isinstance(end_timestamp, float)
                else str(int(end_timestamp.timestamp() * 1000)),
                "jump_channel_id": str(jump_channel_id),
                "remind_type": remind_type
            }
        }

        if description is not None:
            payload["schedule"]["description"] = description

        r = Route('POST', '/channels/{channel_id}/schedules', channel_id=channel_id)
        return self.request(r, json=payload, reason=reason)

    def remove_schedule(self, channel_id: int, schedule_id: int, reason: Optional[str] = None) -> Response[None]:
        r = Route(
            'DELEtE', '/channels/{channel_id}/schedules/{schedule_id}',
            channel_id=channel_id,
            schedule_id=schedule_id
        )

        return self.request(r, reason=reason)

    def mute_member(
            self, user_id: int, guild_id: int, duration: Union[datetime.datetime, int], reason: Optional[str] = None
    ) -> Response[None]:
        payload: Dict[str, Any] = {}
        if isinstance(duration, datetime.datetime):
            payload['mute_end_timestamp'] = str(int(duration.timestamp() * 1000))
        else:
            payload['mute_seconds'] = str(duration)

        r = Route('PATCH', '/guilds/{guild_id}/members/{user_id}/mute', guild_id=guild_id, user_id=user_id)
        return self.request(r, json=payload, reason=reason)

    def mute_members(
            self, user_id: List[int],
            guild_id: int,
            duration: Union[datetime.datetime, int],
            reason: Optional[str] = None
    ) -> Response[None]:
        payload: Dict[str, Any] = {'user_ids': user_id}

        if isinstance(duration, datetime.datetime):
            payload['mute_end_timestamp'] = str(int(duration.timestamp() * 1000))
        else:
            payload['mute_seconds'] = str(duration)

        r = Route('PATCH', '/guilds/{guild_id}/mute', guild_id=guild_id)
        return self.request(r, json=payload, reason=reason)

    def mute_guild(
            self, guild_id: int, duration: Union[datetime.datetime, int], reason: Optional[str] = None
    ) -> Response[None]:
        payload: Dict[str, Any] = {}
        if isinstance(duration, datetime.datetime):
            payload['mute_end_timestamp'] = str(int(duration.timestamp() * 1000))
        else:
            payload['mute_seconds'] = str(duration)

        r = Route('PATCH', '/guilds/{guild_id}/mute', guild_id=guild_id)
        return self.request(r, json=payload, reason=reason)

    def delete_message(
            self, channel_id: int, message_id: str, *, reason: Optional[str] = None, hidetip: bool = False
    ) -> Response[None]:
        r = Route('DELETE', '/channels/{channel_id}/messages/{message_id}', channel_id=channel_id,
                  message_id=message_id)
        params = {'hidetip': 'true' if hidetip else 'false'}
        return self.request(r, reason=reason, params=params)

    def delete_messages(
            self, channel_id: int, message_ids: List[str], *, reason: Optional[str] = None
    ) -> Response[None]:
        r = Route('POST', '/channels/{channel_id}/messages/bulk-delete', channel_id=channel_id)
        payload = {
            'messages': message_ids,
        }
        return self.request(r, json=payload, reason=reason)

    def logs_from(
            self,
            channel_id: int,
            limit: int,
            before: Optional[datetime.datetime] = None,
            after: Optional[datetime.datetime] = None,
            around: Optional[datetime.datetime] = None,
    ) -> Response[List[message.Message]]:
        params: Dict[str, Any] = {
            'limit': limit,
        }

        if before is not None:
            params['before'] = datetime.datetime.timestamp(before)
        if after is not None:
            params['after'] = datetime.datetime.timestamp(after)
        if around is not None:
            params['around'] = datetime.datetime.timestamp(around)

        return self.request(Route('GET', '/channels/{channel_id}/messages', channel_id=channel_id), params=params)

    def get_permission(self, guild_id: int) -> Response[List[PermissionPayload]]:
        return self.request(Route('GET', '/guilds/{guild_id}/api_permission', guild_id=guild_id))

    def demand_permission(
            self,
            guild_id: int,
            channel_id: int,
            desc: str,
            path: str,
            method: str
    ) -> Response[PermissionDemandPayload]:

        payload: Dict[str, Any] = {
            "channel_id": str(channel_id),
            "api_identify": {"path": path, "method": method},
            "desc": desc
        }

        return self.request(Route('POST', '/guilds/{guild_id}/api_permission/demand', guild_id=guild_id), json=payload)

    def ack_interaction(self, interaction_id: str, code: int):
        r = Route(
            'PUT',
            '/interactions/{interaction_id}',
            interaction_id=interaction_id
        )
        payload = {
            "code": code
        }
        return self.request(
            r,
            json=payload
        )

    def global_pin_message(
            self,
            guild_id: int,
            channel_id: Optional[int] = None,
            message_id: Optional[str] = None,
            announces_type: Optional[int] = None,
            recommend_channels: Optional[Dict[int, str]] = None,
            reason: Optional[str] = None
    ):
        r = Route(
            'POST',
            '/guilds/{guild_id}/announces',
            guild_id=guild_id,
        )
        if recommend_channels:
            recommend_channels: List[Dict[str, str]] = [
                {
                    'channel_id': str(m),
                    'introduce': n
                } for m, n in recommend_channels.items()
            ]
        payload: Dict[str, Any] = {}
        if channel_id is not None:
            payload['channel_id'] = channel_id
        if message_id is not None:
            payload['message_id'] = message_id
        if announces_type is not None:
            payload['announces_types'] = announces_type
        if recommend_channels is not None:
            payload['recommend_channels'] = recommend_channels
        return self.request(r, json=payload, reason=reason)

    def global_unpin_message(self, guild_id: int, message_id: str, reason: Optional[str] = None):
        r = Route(
            'DELETE',
            '/guilds/{guild_id}/announces/{message_id}',
            guild_id=guild_id,
            message_id=message_id
        )
        return self.request(r, reason=reason)

    def channel_pin_message(self, channel_id: int, message_id: str, reason: Optional[str] = None):
        r = Route(
            'PUT',
            '/channels/{channel_id}/pins/{message_id}',
            channel_id=channel_id,
            message_id=message_id,
        )
        return self.request(r, reason=reason)

    def channel_unpin_message(self, channel_id: int, message_id: str, reason: Optional[str] = None):
        r = Route(
            'DELETE',
            '/channels/{channel_id}/pins/{message_id}',
            channel_id=channel_id,
            message_id=message_id
        )
        return self.request(r, reason=reason)

    def add_reaction(
            self, channel_id: int, message_id: str, custom: bool, id: int
    ) -> Response[None]:
        r = Route(
            'PUT',
            '/channels/{channel_id}/messages/{message_id}/reactions/{type}/{id}',
            channel_id=channel_id,
            message_id=message_id,
            type=1 if custom else 2,
            id=id
        )
        return self.request(r)

    def remove_reaction(
            self, channel_id: int, message_id: str, custom: bool, id: int
    ) -> Response[None]:
        r = Route(
            'DELETE',
            '/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/{member_id}',
            channel_id=channel_id,
            message_id=message_id,
            type=1 if custom else 2,
            id=id
        )
        return self.request(r)

    def send_guide(
            self,
            channel_id: int,
            content: str,
    ) -> Response[None]:
        r = Route(
            'POST',
            '/channels/{channel_id}/settingguide',
            channel_id=channel_id
        )
        payload = {'content': content}
        return self.request(r, json=payload)

    def get_thread_list(
            self,
            channel_id: int,
    ) -> Response[ThreadListResponse]:
        return self.request(Route('GET', '/channels/{channel_id}/threads', channel_id=channel_id))

    def get_thread(
            self,
            channel_id: int,
            thread_id: int,
    ) -> Response[ThreadPayload]:
        return self.request(Route('GET', '/channels/{channel_id}/threads/{thread_id}', channel_id=channel_id, thread_id=thread_id))

    def create_thread(
            self,
            channel_id: int,
            title: str,
            content: str,
            thread_type: int,
    ) -> Response[ThreadPayload]:
        payload = {
            'title': title,
            'content': content,
            'format': thread_type,
        }
        return self.request(Route('PUT', '/channels/{channel_id}/threads', channel_id=channel_id), json=payload)

    def delete_thread(
            self,
            channel_id: int,
            thread_id: int,
    ) -> Response[None]:
        return self.request(Route('DELETE', '/channels/{channel_id}/threads/{thread_id}', channel_id=channel_id, thread_id=thread_id))



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from typing import Literal, TypedDict, Optional, Union

from .user import PartialUser

ChannelType = Literal[0, 1, 2, 3, 4, 10005, 10006, 10007]


class _BaseChannel(TypedDict):
    id: str
    name: str


class _BaseGuildChannel(_BaseChannel):
    guild_id: str
    position: int
    parent_id: Optional[str]
    owner_id: str
    private_type: int


class DMChannel(_BaseChannel):
    channel_id: str
    type: Literal[1]
    guild_id: str
    create_time: str
    last_message_id: Optional[str]
    recipients: PartialUser


class TextChannel(_BaseGuildChannel):
    type: Literal[0]
    sub_type: str


class LiveChannel(_BaseGuildChannel):
    type: Literal[5]


class VoiceChannel(_BaseGuildChannel):
    type: Literal[2]


class CategoryChannel(_BaseGuildChannel):
    type: Literal[4]


class AppChannel(_BaseGuildChannel):
    type: Literal[6]


class ThreadChannel(_BaseGuildChannel):
    type: Literal[7]


GuildChannel = Union[TextChannel, LiveChannel, VoiceChannel, CategoryChannel, AppChannel, ThreadChannel]
Channel = Union[GuildChannel, DMChannel]



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import asyncio
import datetime
from typing import AsyncIterator, TypeVar, Awaitable, Any, Optional, Callable, Union, List, TYPE_CHECKING

from .error import NoMoreItems
from .object import Object
from .utils import maybe_coroutine

if TYPE_CHECKING:
    from .types.guild import (
        Guild as GuildPayload,
    )
    from .types.message import (
        Message as MessagePayload,
    )
    from .member import Member
    from .message import Message
    from .guild import Guild

__all__ = (
    'GuildIterator',
    'MemberIterator',
)

T = TypeVar('T')
OT = TypeVar('OT')
_Func = Callable[[T], Union[OT, Awaitable[OT]]]


class _AsyncIterator(AsyncIterator[T]):
    __slots__ = ()

    async def next(self) -> T:
        raise NotImplementedError

    def get(self, **attrs: Any) -> Awaitable[Optional[T]]:
        def predicate(elem: T):
            for attr, val in attrs.items():
                nested = attr.split('__')
                obj = elem
                for attribute in nested:
                    obj = getattr(obj, attribute)

                if obj != val:
                    return False
            return True

        return self.find(predicate)

    async def find(self, predicate: _Func[T, bool]) -> Optional[T]:
        while True:
            try:
                elem = await self.next()
            except NoMoreItems:
                return None

            ret = await maybe_coroutine(predicate, elem)
            if ret:
                return elem

    def chunk(self, max_size: int):
        if max_size <= 0:
            raise ValueError('async iterator chunk sizes must be greater than 0.')
        return _ChunkedAsyncIterator(self, max_size)

    def map(self, func: _Func[T, OT]):
        return _MappedAsyncIterator(self, func)

    def filter(self, predicate: _Func[T, bool]):
        return _FilteredAsyncIterator(self, predicate)

    async def flatten(self) -> List[T]:
        return [element async for element in self]

    async def __anext__(self) -> T:
        try:
            return await self.next()
        except NoMoreItems:
            raise StopAsyncIteration()


class _ChunkedAsyncIterator(_AsyncIterator[List[T]]):
    def __init__(self, iterator, max_size):
        self.iterator = iterator
        self.max_size = max_size

    async def next(self) -> List[T]:
        ret: List[T] = []
        n = 0
        while n < self.max_size:
            try:
                item = await self.iterator.next()
            except NoMoreItems:
                if ret:
                    return ret
                raise
            else:
                ret.append(item)
                n += 1
        return ret


class _MappedAsyncIterator(_AsyncIterator[T]):
    def __init__(self, iterator, func):
        self.iterator = iterator
        self.func = func

    async def next(self) -> T:
        # this raises NoMoreItems and will propagate appropriately
        item = await self.iterator.next()
        return await maybe_coroutine(self.func, item)


def _identity(x):
    return x


class _FilteredAsyncIterator(_AsyncIterator[T]):
    def __init__(self, iterator, predicate):
        self.iterator = iterator

        if predicate is None:
            predicate = _identity

        self.predicate = predicate

    async def next(self) -> T:
        getter = self.iterator.next
        pred = self.predicate
        while True:
            # propagate NoMoreItems similar to _MappedAsyncIterator
            item = await getter()
            ret = await maybe_coroutine(pred, item)
            if ret:
                return item


class GuildIterator(_AsyncIterator['Guild']):
    def __init__(self, bot, limit, before=None, after=None):

        self.bot = bot
        self.limit = limit
        self.before = before
        self.after = after

        self._filter = None
        self.state = self.bot._connection

        self.get_guilds = self.bot.http.get_guilds
        self.get_guild_channels = self.bot.http.get_guild_channels
        self.guilds = asyncio.Queue()

        if self.before and self.after:
            self._retrieve_guilds = self._retrieve_guilds_before_strategy  # type: ignore
            self._filter = lambda m: int(m['id']) > self.after.id
        elif self.after:
            self._retrieve_guilds = self._retrieve_guilds_after_strategy  # type: ignore
        else:
            self._retrieve_guilds = self._retrieve_guilds_before_strategy  # type: ignore

    async def next(self) -> Guild:
        if self.guilds.empty():
            await self.fill_guilds()

        try:
            return self.guilds.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems()

    def _get_retrieve(self):
        l = self.limit
        if l is None or l > 100:
            r = 100
        else:
            r = l
        self.retrieve = r
        return r > 0

    def create_guild(self, data):
        from .guild import Guild

        return Guild(data=data, state=self.state)

    async def fill_guilds(self):
        if self._get_retrieve():
            data = await self._retrieve_guilds(self.retrieve)
            if self.limit is None or len(data) < 100:
                self.limit = 0

            if self._filter:
                data = filter(self._filter, data)

            for element in data:
                await self.guilds.put(self.create_guild(element))

    async def _retrieve_guilds(self, retrieve) -> List[Guild]:
        """Retrieve guilds and update next parameters."""
        raise NotImplementedError

    async def _retrieve_guilds_before_strategy(self, retrieve):
        """Retrieve guilds using before parameter."""
        before = self.before.id if self.before else None
        data: List[GuildPayload] = await self.get_guilds(retrieve, before=before)
        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
        return data

    async def _retrieve_guilds_after_strategy(self, retrieve):
        """Retrieve guilds using after parameter."""
        after = self.after.id if self.after else None
        data: List[GuildPayload] = await self.get_guilds(retrieve, after=after)
        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
        return data


class MemberIterator(_AsyncIterator['Member']):
    def __init__(self, guild, limit=1000, after=None):

        if isinstance(after, int):
            after = Object(id=after)

        self.guild = guild
        self.limit = limit
        self.after = after or 0

        self.state = self.guild._state
        self.get_members = self.state.http.get_members
        self.members = asyncio.Queue()

    async def next(self) -> Member:
        if self.members.empty():
            await self.fill_members()

        try:
            return self.members.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems()

    def _get_retrieve(self):
        l = self.limit
        if l is None or l > 400:
            r = 400
        else:
            r = l
        self.retrieve = r
        return r > 0

    async def fill_members(self):
        if self._get_retrieve():
            after = self.after.id if self.after else None
            data = await self.get_members(self.guild.id, self.retrieve, after)
            if not data:
                # no data, terminate
                return

            self.after = Object(id=int(data[-1]['user']['id']))

            for element in reversed(data):
                await self.members.put(self.create_member(element))

    def create_member(self, data):
        from .member import Member

        return Member(data=data, guild=self.guild, state=self.state)


class HistoryIterator(_AsyncIterator['Message']):
    """用于接收频道消息历史的迭代器。
    消息端点有两个我们关心的行为：
    如果指定了 ``before`` ，则消息端点返回 ``before`` 之前的 ``limit`` 最新消息，以最新的优先排序。
    要填充超过 100 条消息，请将 ``before`` 参数更新为收到的最旧消息。消息将按时间顺序返回。
    如果指定了 ``after``，它返回 ``after`` 之后的 ``limit`` 最旧的消息，以最新的在前排序。
    要填充超过 100 条消息，请将 ``after`` 参数更新为收到的最新消息。如果消息没有反转，它们将乱序（99-0、199-100 等）
    注意如果同时指定了 ``before`` 和 ``after`` ，则 ``before`` 将被忽略。

    Parameters
    -----------
    messageable: :class:`abc.Messageable`
        可从中检索消息历史记录的 Messageable 类。

    limit: :class:`int`
        要检索的最大消息数

    before: Optional[:class:`datetime.datetime`]
        所有消息必须在其之前的消息。

    after: Optional[:class:`datetime.datetime`]
        所有消息必须在其后的消息。

    around: Optional[:class:`datetime.datetime`]
        所有消息必须围绕的消息。 Limit max 101。注意，如果limit是偶数，这将最多返回limit+1条消息。

    oldest_first: Optional[:class:`bool`]
        如果设置为 ``True``，以最旧->最新的顺序返回消息。如果指定了“after”，则默认为“True”，否则为“False”。
    """

    def __init__(self, messageable, limit, before=None, after=None, around=None, oldest_first=None):

        if oldest_first is None:
            self.reverse = after is not None
        else:
            self.reverse = oldest_first

        self.messageable = messageable
        self.limit = limit
        self.before = before
        self.after = after or 0
        self.around = around

        self._filter = None  # message dict -> bool

        self.state = self.messageable._state
        self.logs_from = self.state.http.logs_from
        self.messages = asyncio.Queue()

        if self.around:
            if self.limit is None:
                raise ValueError('历史不支持limit=None')
            if self.limit > 101:
                raise ValueError("指定 around 参数时的历史最大限制 101")
            elif self.limit == 101:
                self.limit = 100  # Thanks qq

            self._retrieve_messages = self._retrieve_messages_around_strategy  # type: ignore
            if self.before and self.after:
                self._filter = lambda m: \
                    self.timestamp(self.after) < self.timestamp(m['timestamp']) < self.timestamp(
                        self.before)  # type: ignore
            elif self.before:
                self._filter = lambda m: self.timestamp(m['timestamp']) < self.timestamp(self.before)  # type: ignore
            elif self.after:
                self._filter = lambda m: self.timestamp(self.after) < self.timestamp(m['timestamp'])  # type: ignore
        else:
            if self.reverse:
                self._retrieve_messages = self._retrieve_messages_after_strategy  # type: ignore
                if self.before:
                    self._filter = lambda m: self.timestamp(m['timestamp']) < self.timestamp(
                        self.before)  # type: ignore
            else:
                self._retrieve_messages = self._retrieve_messages_before_strategy  # type: ignore
                if self.after and self.after != 0:
                    self._filter = lambda m: self.timestamp(m['timestamp']) > self.timestamp(self.after)  # type: ignore

    def timestamp(self, dt: Union[datetime.datetime, str]):
        if isinstance(dt, str):
            dt = datetime.datetime.strptime(dt, "%Y-%m-%dT%H:%M:%SZ")
        return int(datetime.datetime.timestamp(dt))

    async def next(self) -> Message:
        if self.messages.empty():
            await self.fill_messages()

        try:
            return self.messages.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems()

    def _get_retrieve(self):
        l = self.limit
        if l is None or l > 100:
            r = 100
        else:
            r = l
        self.retrieve = r
        return r > 0

    async def fill_messages(self):
        if not hasattr(self, 'channel'):
            # do the required set up
            channel, private = await self.messageable._get_channel()
            self.channel = channel

        if self._get_retrieve():
            data = await self._retrieve_messages(self.retrieve)
            if len(data) < 100:
                self.limit = 0  # terminate the infinite loop

            if self.reverse:
                data = reversed(data)
            if self._filter:
                data = filter(self._filter, data)

            channel = self.channel
            for element in data:
                await self.messages.put(self.state.create_message(channel=channel, data=element))

    async def _retrieve_messages(self, retrieve) -> List[Message]:
        """检索消息并更新下一个参数"""
        raise NotImplementedError

    async def _retrieve_messages_before_strategy(self, retrieve):
        """使用 before 参数检索消息。"""
        before = self.timestamp(self.before) if self.before else None
        data: List[MessagePayload] = await self.logs_from(self.channel.id, retrieve, before=before)
        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            self.before = datetime.datetime.strptime(data[-1]['timestamp'], "%Y-%m-%dT%H:%M:%SZ")
        return data

    async def _retrieve_messages_after_strategy(self, retrieve):
        """使用 after 参数检索消息。"""
        after = self.timestamp(self.after) if self.after else None
        data: List[MessagePayload] = await self.logs_from(self.channel.id, retrieve, after=after)
        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            self.after = datetime.datetime.strptime(data[0]['timestamp'], "%Y-%m-%dT%H:%M:%SZ")
        return data

    async def _retrieve_messages_around_strategy(self, retrieve):
        """使用 around 参数检索消息。"""
        if self.around:
            around = self.around.id if self.around else None
            data: List[MessagePayload] = await self.logs_from(self.channel.id, retrieve, around=around)
            self.around = None
            return data
        return []



from setuptools import setup

from setuptools import setup
import re

with open('requirements.txt', encoding='utf-8') as f:
  requirements = f.read().splitlines()

with open('qq/__init__.py', encoding='utf-8') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('version is not set')

if version.endswith(('a', 'b', 'rc')):
    # append version identifier based on commit count
    try:
        import subprocess
        p = subprocess.Popen(['git', 'rev-list', '--count', 'HEAD'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            version += out.decode('utf-8').strip()
        p = subprocess.Popen(['git', 'rev-parse', '--short', 'HEAD'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            version += '+g' + out.decode('utf-8').strip()
    except Exception:
        pass

with open('README.md', encoding='utf-8') as f:
    readme = f.read()

extras_require = {
    'docs': [
        'sphinx==4.0.2',
        'sphinxcontrib_trio==1.1.2',
        'sphinxcontrib-websupport',
    ],
}

setup(
    name='qq.py',
    version=version,
    description='QQ 频道 API 的 Python Wrapper',
    py_modules=["qq"],
    packages=['qq', "qq.types", "qq.ext.commands", "qq.ext.tasks"],
    license='MIT',
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    extras_require=extras_require,
    python_requires='>=3.8.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Natural Language :: Chinese (Simplified)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Framework :: AsyncIO',
        'Framework :: aiohttp',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Typing :: Typed',
    ],
    url="https://github.com/foxwhite25/qq.py",
    author='foxwhite25',
    author_email='vct.xie@gmail.com'
)



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .channel import ThreadChannel
    from .state import ConnectionState
    from .types.forum import Thread as ThreadPayload, ThreadInfo

__all__ = (
    'Thread',
)


class Thread:
    """代表一个 QQ 帖子。

    Attributes
    -----------
    author: :class:`User`
        帖子的作者。
    guild: :class:`Guild`
        帖子所在的服务器。
    channel: :class:`ThreadChannel`
        帖子所在的频道。
    id: :class:`int`
        帖子 ID。
    title: :class:`str`
        帖子标题。
    content: :class:`str`
        帖子内容。
    created_at: :class:`datetime.datetime`
        发布时间。
    """

    __slots__ = (
        '_state',
        '_author',
        '_guild',
        '_channel',
        'author',
        'guild',
        'channel',
        'id',
        'title',
        'content',
        'created_at',
    )

    def __init__(
            self,
            *,
            state: ConnectionState,
            channel: ThreadChannel,
            data: ThreadPayload,
    ):
        self._state = state
        self._channel = int(data['channel_id'])
        self._author = int(data['author_id'])
        self._guild = int(data['guild_id'])
        self.channel = channel
        self.author = channel.guild.get_member(self._author)
        self.guild = self._state._get_guild(self._guild)
        self._update(data['thread_info'])

    def _update(self, data: ThreadInfo):
        self.id = int(data['thread_id'])
        self.title = data['title']
        self.content = data['content']
        self.created_at = datetime.fromtimestamp(data['datetime'])

    async def fetch(self):
        """从服务器获取最新的帖子数据。"""
        data = await self._state.http.get_thread(self.channel.id, self.id)
        self._update(data)

    async def delete(self):
        """删除帖子。"""
        await self._state.http.delete_thread(self.channel.id, self.id)



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import re
from typing import Any, Dict, Optional, TYPE_CHECKING, Type, TypeVar, Union, Tuple

from .asset import AssetMixin
from .error import InvalidArgument

__all__ = (
    'PartialEmoji',
)

if TYPE_CHECKING:
    from .state import ConnectionState
    from .types.message import PartialEmoji as PartialEmojiPayload


class _EmojiTag:
    __slots__ = ()

    id: int

    def _to_partial(self) -> PartialEmoji:
        raise NotImplementedError


PE = TypeVar('PE', bound='PartialEmoji')


class PartialEmoji(_EmojiTag, AssetMixin):
    """代表 “部分” 表情符号。
    该模型将在两种情况下给出:

    - “原始”数据事件，例如 :func:`on_raw_reaction_add`
    - 机器人无法看到的自定义表情符号，例如 :attr:`Message.reactions`

    .. container:: operations

        .. describe:: x == y

            检查两个表情符号是否相同。

        .. describe:: x != y

            检查两个表情符号是否不同。

        .. describe:: hash(x)

            返回表情符号的哈希值。

        .. describe:: str(x)

            返回为 QQ 渲染的表情符号。


    Attributes
    -----------
    custom: :class:`bool`
        表情是否是 QQ 自定义表情。
    id: :class:`int`
        自定义表情符号的 ID（如果适用）。
    """

    __slots__ = ('animated', 'name', 'id', '_state', 'custom')

    _CUSTOM_EMOJI_RE = re.compile(r'<?emoji:(?P<id>[0-9]{13,20})>?')

    if TYPE_CHECKING:
        id: Optional[int]

    def __init__(self, *, custom: bool, id: str = None):
        self.custom = custom
        self.id = int(id)
        self._state: Optional[ConnectionState] = None
        if not custom:
            self.name = chr(self.id)

    @classmethod
    def from_dict(cls: Type[PE], data: Union[PartialEmojiPayload, Dict[str, Any]]) -> PE:
        return cls(
            id=int(data.get('id')),
            custom=True if data.get('type') == 1 else False,
        )

    @classmethod
    def from_str(cls: Type[PE], value: str) -> PE:
        """将表情符号的 QQ 字符串表示形式转换为 :class:`PartialEmoji`。
        接受的格式是：

        - ``emoji:id``
        - ``<emoji:id>``

        如果格式不匹配，则假定它是一个 unicode 表情符号，取第一个字符作为 emoji。

        Parameters
        ------------
        value: :class:`str`
            表情符号的字符串表示。

        Returns
        --------
        :class:`PartialEmoji`
            此字符串中的表情符号。
        """
        match = cls._CUSTOM_EMOJI_RE.match(value)
        if match is not None:
            groups = match.groupdict()
            emoji_id = groups['id']
            return cls(id=int(emoji_id), custom=True)
        value = ord(value[0])
        return cls(id=value, custom=False)

    def to_dict(self) -> Dict[str, Any]:
        o: Dict[str, Any] = {'id': self.id, 'type': '1' if self.custom else '2'}
        return o

    def _to_partial(self) -> PartialEmoji:
        return self

    @classmethod
    def with_state(
            cls: Type[PE], state: ConnectionState, *, custom: bool, id: str = None
    ) -> PE:
        self = cls(custom=custom, id=id)
        self._state = state
        return self

    def __str__(self) -> str:
        if self.id is None:
            return self.name
        if self.animated:
            return f'<a:{self.name}:{self.id}>'
        return f'<{self.name}:{self.id}>'

    def __repr__(self):
        return f'<{self.__class__.__name__} id={self.id} type={"1" if self.custom else "2"}>'

    def __eq__(self, other: Any) -> bool:
        if self.is_unicode_emoji():
            return isinstance(other, PartialEmoji) and self.id == other.id

        if isinstance(other, _EmojiTag):
            return self.id == other.id
        return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((self.id, self.name))

    def is_custom_emoji(self) -> bool:
        """:class:`bool`: 检查这是否是自定义的非 Unicode 表情符号。"""
        return self.custom

    def is_unicode_emoji(self) -> bool:
        """:class:`bool`: 检查这是否是 Unicode 表情符号。"""
        return not self.custom

    def _as_reaction(self) -> Tuple[bool, int]:
        if self.id is None:
            return self.name
        return self.custom, self.id

    async def read(self) -> bytes:
        if self.is_unicode_emoji():
            raise InvalidArgument('PartialEmoji 不是自定义表情符号')

        return await super().read()

    @property
    def to_string(self):
        return f'<emoji:{self.id}>' if self.custom else chr(self.id)
